"""
 Copyright 2015 Cristofor Nogueira Gascons
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import xmlWritter

__author__ = 'cris nogueira'

from model import ModelXHSTT
from xmlparser import Parser
from xmlWritter import XMLWritter
import time
import sys, os
import math
import itertools
from subprocess import PIPE, Popen

class ModelSAT(ModelXHSTT):

    def __init__(self, xml_source, mto=False, timeout=1800000):
        ModelXHSTT.__init__(self)

        # Variable declaration
        self.Xt = dict()            # event times
        self.Xs = dict()            # event starting times
        self.Xd = dict()            # event durations
        self.Xr = dict()            # dictionary containing, for each resource all the events that demand such resource

        if mto:
            self.card = ModuloTotalizer()
            print 'Using MTO encoding'
        else:
            self.card = MCardinalityNetwork(5)
            print 'Using Mixed Cardinality Network encoding'

        self.clauses = []

        Parser(self, xml_source).parse_model()

        # remove all duplicated clauses
        clauses = set()
        for clause in self.clauses:
            if clause.is_empty():
                print 'Empty clause found!!'
            else:
                clauses.add(clause)
        self.clauses = list(clauses)

        print 'parsing stage complete'

        print 'all set, starting check'


        # flush to file
        filename = xml_source[xml_source.find('/')+1:xml_source.find('.')]
        filepath = os.getcwd() + '/output/' + filename + '.wcnf'
        print filepath
        dimacs_file = open(filepath, 'w')
        dimacs_file.write('c\nc File generated by ModelSAT.py\nc\n')
        dimacs_file.write('p wcnf %i %i %i\n' % (Boolean.NUM_VARS, len(self.clauses), Clause.TOP))
        for i, clause in enumerate(self.clauses):
            dimacs_file.write(str(clause) + '\n')
            if (i+1)%10000 == 0:
              dimacs_file.flush()
        dimacs_file.close()

        tic = time.time()
        # Invoke the MAXSAT solver
        # process = Popen(['./sat-solvers/open-wbo/open-wbo', '-cardinality=2', filepath], stdout=PIPE)
        # process = Popen(['./sat-solvers/open-wbo/open-wbo', '-cardinality=2', filepath], stdout=PIPE)
        # process = Popen(['./sat-solvers/mscg/mscg-linux-x86-64',filepath], stdout=PIPE)
        # process = Popen(['./sat-solvers/msuncore/msuncore', '-r','bcd2', '-hard', filepath], stdout=PIPE)
        # process = Popen(['cat','output/out.dimacs'], stdout=PIPE)
        process = Popen(['./sat-solvers/pwbo/pwbo', '-thread=4','-time-limit='+str(timeout), filepath], stdout=PIPE)
        # process = Popen(['java', '-jar', 'sat-solvers/sat4j/sat4j-maxsat.jar','-T', str(timeout), '-k', filepath], stdout=PIPE)
        stdout, stderr = process.communicate()

        print ''
        print '###############################################'
        print ''


        for line in stdout.split('\n'):
            if line.startswith('v'):
                self.solution = set()
                for elem in line[2:].split(' '):
                    if len(elem) > 0 and int(elem) > 0:
                        self.solution.add(Boolean(int(elem)))
            else:
                print line

        if len(self.solution) > 0:
            xmlWritter.XMLWritter(xml_source, self)

        print 'took ', time.time()-tic, ' seconds'

    def get_times_assignments(self):
        time_count = len(self.times)
        res = []
        if self.solution is not None:
            for event in self.events.itervalues():
                for d in range(1, event.duration+1):
                    variables = self.Xd[event.num][d]

                    # for each set variable, create an entry
                    for i in range(0, time_count):
                        if variables[i] in self.solution:
                            res.append((event.identifier, d, self.num2time[i].identifier))


        return res

    def onParsedEvents(self, spread_constraint=True):
        """
        For each parsed event, declare all related variables. that is time and resource variables
        alongside its known domains (albeit some restrictions still to be known such as prefer-like constraints)

        Note that on this stage, all event object contains only the needed declarations. That is in the extreme case
        that any event would have any role declared without any resource referenced nor assign_resource constraint,
        such role would only appear in the xml and not in it's object representation. Thus all events contains only roles
        for which either have a preassigned resource or solver needs to assign one to it.
        """
        Xt = self.Xt
        Xs = self.Xs
        Xd = self.Xd

        # initialize Xr
        Xr = self.Xr
        for resource in self.resources.itervalues():
            Xr[resource.num] = set()

        time_count = len(self.times)
        counter = 0
        sys.stdout.write('Parsing events ...')
        constraints = []
        for event in self.events.itervalues():
            counter += 1
            sys.stdout.write('\rParsing events\t\t%i/%i' % (counter, len(self.events)))
            e = event.num

            # time variables
            Xt[e] = BooleanVec(time_count)
            Xs[e] = BooleanVec(time_count)

            # preassigned times
            if event.hasPreassignedTime():
                if event.duration != 1:
                    raise Exception('Encountered an Event with preassigned time with a duration greater than 1')

                preassigned = self.times[event.time].num
                args = [Clause(~Xt[e][t]) for t in range(time_count) if t != preassigned]
                args.append(Clause(Xt[e][preassigned]))
                constraints.extend(args)
            else:
                constraints.extend(self.card.exactly_k(Xt[e], event.duration))
                # Xs is contained within Xt
                constraints.extend([Clause([~Xs[e][t], Xt[e][t]]) for t in range(time_count)])  # Xs implies Xt
                # If Event is tanking place in Xt and not in Xt-1 then it must be a starting time
                constraints.extend([Clause([~Xt[e][t], Xt[e][t-1], Xs[e][t]]) for t in range(1, time_count)])
                # If an event occurs at the first time slot (a.k.a. num 0) then it must be a starting time also
                constraints.append(Clause([~Xt[e][0], Xs[e][0]]))

            # Xd constraints:
            #   i -> stands for the BitVector representing the starting times with duration 'i'
            Xd[e] = dict()
            for i in range(1, event.duration+1):
                Xd[e][i] = BooleanVec(time_count)
                constraints.extend([Clause([~Xd[e][i][t], Xs[e][t]]) for t in range(time_count)])  # Xd implies Xs
                if not spread_constraint:
                    # The time immidiately after an event of duration d, must be set to false
                    constraints.extend([Clause([~Xd[e][i][t], ~Xt[e][t+i]]) for t in range(time_count-i)])
                    # Also negate all unfeasible times (the last ones)
                    constraints.extend([Clause([~Xd[e][i][t]]) for t in range(time_count-i+1, time_count)])
                # When a bit of Xd[e][i] is set, ensure that the event takes place in 'i' consecutive slots
                constraints.extend([Clause([~Xd[e][i][t], Xt[e][ti]])
                                    for t in range(time_count-i+1) for ti in range(t, t+i)])
                # If an event with duration d is taking place, make sure no other events start while
                # the event is taking place (i.e. within the duration range)
                if i > 1:
                    constraints.extend([Clause([~Xd[e][i][t], ~Xs[e][ti]])
                                        for t in range(time_count-i+1) for ti in range(t+1, t+i)])
                    constraints.extend([Clause([~Xd[e][i][t], ~Xd[e][di][ti]])
                                        for t in range(time_count-i+1) for di in range(1, i) for ti in range(t, t+i)])


            # If an event starts at time t, then exactly one Xd must be set
            if event.duration > 1:
                for t in range(time_count):
                    # If a duration is set for a specific time, make sure no other durations are set on that time
                    args = [Xd[e][i][t] for i in range(1, event.duration+1)]
                    args.append(~Xs[e][t])
                    constraints.extend(self.card.exactly_k(args, 1))
            else:
                constraints.extend([Clause([~Xs[e][t], Xd[e][1][t]]) for t in range(time_count)])

            # resource variables (this model requires all resources to be preassigned)
            # check that the event has all resources preassigned
            preassigned_resources = event.get_preassigned_resources_nums()
            if len(preassigned_resources) < len(event.resources):
                raise Exception('At least one event with one resource left to assign')

            # Add event reference into all preassigned resources entries in Xr
            for res_num in preassigned_resources:
                Xr[res_num].add(e)

        sys.stdout.write('\n')
        self.clauses.extend(constraints)

    def set_time_for_event(self, event_num, duration, start_t):
        #self.clauses.append(Clause([self.Xd[event_num][duration][start_t]]))
        pass

    def avoidClashesConstraint(self, cost, resources_ids):

        i = 0
        sys.stdout.write('Asserting Avoid Clashes\t\t%i/%i' % (i, len(resources_ids)))
        time_count = len(self.times)
        Xt = self.Xt
        weight = cost/len(resources_ids)
        for resource_id in resources_ids:

            i += 1
            sys.stdout.write('\rAsserting Avoid Clashes\t\t%i/%i' % (i, len(resources_ids)))
            resource = self.resources[resource_id]
            requiring_events = self.Xr[resource.num]

            for t in range(time_count):
                self.clauses.extend(weight_clauses(self.card.at_most_1([Xt[e][t] for e in requiring_events]), weight))

        sys.stdout.write('\n')

    def preferTimesConstraint(self, cost, events_ids, times_ids, duration):

        sys.stdout.write('Asserting Prefer Times constraints')
        time_count = len(self.times)

        # Create a set with all forbidden time slots
        forbidden = set(range(time_count))
        for time_id in times_ids:
            forbidden.remove(self.times[time_id].num)

        weight = cost/len(events_ids)
        p_weight = max(1, weight/3)  # partial weight

        # for each event, forbid all forbidden times for the given duration (if any)
        for event_id in events_ids:
            constraints = []
            event = self.events[event_id]
            e = event.num
            if duration is None:
                for d in range(1, event.duration+1):
                    constraints.extend([Clause([~self.Xd[e][d][t]]) for t in forbidden])
                constraints.extend([Clause([~self.Xt[e][t]]) for t in forbidden])   # Only neeeded constraint is this one (<--)
                constraints.extend([Clause([~self.Xs[e][t]]) for t in forbidden])   # All others are redundant
            elif duration in self.Xd[e]:
                constraints.extend([Clause([~self.Xd[e][duration][t]]) for t in forbidden])

            self.clauses.extend(weight_clauses(constraints, weight))
        sys.stdout.write('\t\t[DONE]\n')

    def splitEventsConstraint(self, cost, events_id, minimum, maximum, min_amount, max_amount):

        sys.stdout.write('Split Events constraints')

        Xs = self.Xs
        Xd = self.Xd
        time_count = len(self.times)
        constraints = []

        weight = max(1, cost/len(events_id))

        for event_id in events_id:
            event = self.events[event_id]
            e = event.num

            # nullify all durations below minimum
            for d in range(1, minimum):
                constraints.extend([Clause([~Xd[e][d][t]]) for t in range(time_count)])

            # nullify all durations above maximum
            for d in range(maximum+1, event.duration+1):
                constraints.extend([Clause([~Xd[e][d][t]]) for t in range(time_count)])

            if min_amount == max_amount and min_amount > 0:
                constraints.extend(self.card.exactly_k(Xs[e], min_amount))
            else:
                if min_amount > 0:
                    constraints.extend(self.card.at_least_k(Xs[e], min_amount))

                if max_amount < time_count:
                    constraints.extend(self.card.at_most_k(Xs[e], max_amount))

        self.clauses.extend(weight_clauses(constraints, weight))
        sys.stdout.write('\t\t\t\t[DONE]\n')

    def spreadEventsConstraint(self, cost, event_groups, time_groups_ids):
        sys.stdout.write('Spread Events constraints')

        Xs = self.Xs
        Xt = self.Xt
        Xd = self.Xd
        time_count = len(self.times)

        weight = max(1, cost/len(event_groups))

        time_ranges = []

        for time_group_id, min_max in time_groups_ids.iteritems():
            time_group = self.time_groups[time_group_id]
            minimum = min_max[0]
            maximum = min_max[1]
            time_nums = []
            for time_id in time_group.elems:
                time_nums.append(self.times[time_id].num)
            time_nums.sort()
            # Generate a list with the time_nums that do not have an immediate predecessor within the group
            # And impose that for those times, if Xt then Xs (that is, those times will always be starting times)
            heads = [t for i, t in enumerate(time_nums) if t != time_nums[i-1]+1]

            # add all consecutive time_ranges into time_ranges
            i = 0
            while i < len(time_nums)-1:
                r = []
                while i < len(time_nums)-1 and time_nums[i] == time_nums[i+1]-1:
                    r.append(time_nums[i])
                    i += 1
                r.append(time_nums[i])
                i += 1
                time_ranges.append(r)


            for event_group_id in event_groups:
                constraints = []
                event_group = self.event_groups[event_group_id]
                # z_args will contain all variables representing any event from event_group starting at an eligible time
                z_args = []  # Xts will have a bit set iff one event of the group has its corresponding bit set in its Xt
                for event_id in event_group.elems:
                    event = self.events[event_id]
                    e = event.num
                    # Impose that for head times, if Xt then Xs (that is, those times will always be starting times)
                    # such that all starting times of the time group are marked as such in Xs if they are set in Xt.
                    # that is, times from which its predecessor time is not within the time group,
                    constraints.extend([Clause([~Xt[e][t], Xs[e][t]]) for t in heads])

                    # fill z_args with Xs variables inside time_group
                    z_args.extend([Xs[e][t] for t in time_nums])

                self.clauses.extend(self.card.weighted_card(z_args, 0, minimum, len(z_args), maximum, weight))


                self.clauses.extend(constraints)

        # time_ranges will contain a list of all consecutive time ranges, so for each event,
        for event in self.events.itervalues():
            e = event.num
            for tr in time_ranges:
                for d in range(1, event.duration+1):
                    # initial, if the event is scheduled with duration d, then time t+d must be free
                    for i in range(0, len(tr)-d):
                        self.clauses.append(Clause([~Xd[e][d][tr[i]], ~Xt[e][tr[i]+d]]))
                        clause = Clause([~Xt[e][t] for t in range(tr[i], tr[i]+d)])
                        clause.add(Xt[e][tr[i]+d])
                        if i > 0:
                            clause.add(Xt[e][tr[i-1]])
                        clause.add(Xd[e][d][tr[i]])
                        self.clauses.append(clause)
                    # final, if all the last d times of a time_range are set, then Xd must be set also
                    if len(tr)-d >= 0:
                        clause = Clause([~Xt[e][t] for t in range(tr[len(tr)-d], tr[len(tr)-1]+1)])
                        if len(tr)-d > 0:
                            clause.add(Xt[e][len(tr)-d-1])
                        clause.add(Xd[e][d][tr[len(tr)-d]])
                        self.clauses.append(clause)
                    # also nullify all unfeasible times
                    ini = max(len(tr)-d+1, 0)
                    self.clauses.extend([Clause([~Xd[e][d][tr[i]]]) for i in range(ini, len(tr))])

        # Fix one teacher
        # for resource in self.resources.itervalues():
        #     if resource.rtype == 'Teacher':
        #         # Impose that the particular teacher must attend on monday, tuesday and/or wednesday
        #         for e in self.Xr[resource.num]:
        #             self.clauses.append(Clause([Xs[e][t] for i in range(min(len(time_ranges), 3))
        #                                         for t in time_ranges[i]]))

        # for resource in self.resources.itervalues():
        #     if resource.rtype == 'Teacher':
        #         event_nums = self.Xr[resource.num]
        #         for e1 in event_nums:
        #             for e2 in event_nums:
        #                 if e1 < e2:
        #                     for tr in time_ranges:
        #                         self.clauses.extend([Clause([~Xs[e2][t], ~Xs[e1][ti]])
        #                                              for t in range(tr[0], tr[-1]) for ti in range(t+1, tr[-1]+1)])



        sys.stdout.write('\t\t\t\t[DONE]\n')

    def avoidUnavailableTimesConstraint(self, cost, resource_ids, time_ids):

        sys.stdout.write('\rAvoid Unavailable times constraints')

        weight = cost/len(resource_ids)

        # Generate a set with all unavailable times numbers
        unavailable_times = set()
        for time_id in time_ids:
            unavailable_times.add(self.times[time_id].num)

        for resource_id in resource_ids:
            # get all Events that the resource are to attend
            event_nums = self.Xr[self.resources[resource_id].num]

            constraints = []
            for e in event_nums:
                constraints.extend([Clause([~self.Xt[e][t]]) for t in unavailable_times])

            self.clauses.extend(weight_clauses(constraints, weight))

        sys.stdout.write('\t\t[DONE]\n')

    def distributeSplitEventsConstraints(self, cost, events_ids, duration, minimum, maximum):

        sys.stdout.write('\rDistribute Split Events constraints')

        weight = max(1, cost/len(events_ids))


        for event_id in events_ids:
            event = self.events[event_id]


            if duration not in self.Xd[event.num]:
                raise Exception('invalid duration on distribute split events constraint for event: '
                                '%s (duration: %i)' % (event_id, duration))

            Xd_vars = [self.Xd[event.num][duration][t] for t in range(len(self.times))]

            #self.clauses.extend(self.card.weighted_card(Xd_vars, 0, minimum, event.duration/duration, maximum, weight))
            # if minimum == maximum and minimum > 0:
            #     constraints.extend(self.card.exactly_k(Xd_vars, minimum))
            # else:
            #     if minimum > 0:
            #         constraints.extend(self.card.at_least_k(Xd_vars, minimum))
            #     if maximum < event.duration:
            #         constraints.extend(self.card.at_most_k(Xd_vars, maximum))

        sys.stdout.write('\t\t[DONE]\n')

    def limitIdleTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):
        
        sys.stdout.write('\rLimit Idle Times constraints')

        weight = max(1, cost/len(resource_ids))

        for resource_id in resource_ids:
            r = self.resources[resource_id].num

            idle_vars = []
            for time_group in time_groups:
                time_ids = self.time_groups[time_group].elems
                time_nums = set()
                for time_id in time_ids:
                    time_nums.add(self.times[time_id].num)

                max_time = max(time_nums)
                min_time = min(time_nums)

                # Check compactness
                for t in range(min_time+1, max_time):
                    if t not in time_nums:
                        raise Exception('Encountered a non compact TimeGroup in LimitIdleTimesConstraint')

                # Get a set with all the eligible times (times eligible to be IDLE inside the time group)
                eligible_times = time_nums
                eligible_times.remove(min_time)
                eligible_times.remove(max_time)

                for t in eligible_times:
                    before = []     # get all vars that represent time_nums that go before this time
                    after = []      # get all vars that represent time_nums that go after this time
                    actual = []        # get all vars that represent time_num 't'
                    for e in self.Xr[r]:
                        before.extend([self.Xt[e][tb] for tb in range(min_time, t)])
                        after.extend([self.Xt[e][ta] for ta in range(t+1, max_time)])
                        actual.append(self.Xt[e][t])

                    constraints = []


                    # for each eligible_time create an auxiliary variable that will remain true iff the time is idle.
                    # Note: To lower the amount of clauses the condition will relax if the minimum equals to 0
                    #       in such case, the auxiliary var will remain true if the time is idle, otherwise unknown.
                    idle = Boolean()
                    # (~act ^ a ^ b) -> idle
                    constraints.extend([Clause([act, ~a, ~b, idle]) for act in actual for b in before for a in after])
                    if minimum > 0:
                        c1 = Clause(after)
                        c2 = Clause(before)
                        # idle implies that there is at least one occupied time in 'before' and in 'after'
                        c1.add(~idle)
                        c2.add(~idle)
                        # also idle implies that all actual times are unoccupied
                        constraints.extend([c1, c2] + [Clause([~idle, ~act]) for act in actual])
                    idle_vars.append(idle)

                    self.clauses.extend(constraints)

            # impose cardinality constraint over idle_vars
            self.clauses.extend(self.card.weighted_card(idle_vars, 0, minimum, len(idle_vars), maximum, weight))

        sys.stdout.write('\t\t\t[DONE]\n')

    def clusterBusyTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):

        sys.stdout.write('\rCluster Busy Times constraint')

        weight = max(1, cost/len(resource_ids))

        time_count = len(self.times)
        resource_nums = set()
        for resource_id in resource_ids:
            resource_nums.add(self.resources[resource_id].num)

        for r in resource_nums:
            constraints = []

            # List containing all auxiliary variables
            busies = []
            for time_group in time_groups:
                time_ids = self.time_groups[time_group].elems
                time_nums = set()
                for time_id in time_ids:
                    time_nums.add(self.times[time_id].num)

                # rt_args contains all Xt variables bounded to 'r'
                rt_args = [self.Xt[e][t] for e in self.Xr[r] for t in time_nums]
                busy = Boolean()
                constraints.append(Clause([~busy] + rt_args))
                constraints.extend([Clause([~xt, busy]) for xt in rt_args])

                # finally append the busy var to the rest
                busies.append(busy)

            # Apply cardinality constraints to busy vars
            constraints.extend(self.card.weighted_card(busies, 0, minimum, len(time_groups), maximum, weight))

            # if minimum == maximum:
            #     constraints.extend(self.card.exactly_k(busies, minimum))
            # else:
            #     if minimum > 0:
            #         constraints.extend(self.card.at_least_k(busies, minimum))
            #     if maximum < len(busies):
            #         constraints.extend(self.card.at_most_k(busies, maximum))

            self.clauses.extend(constraints)

        sys.stdout.write('\t\t\t[DONE]\n')

#########################################
#               Helpers                 #
#########################################

def simplify_clauses(clauses):
    """
    fixed point algorithm that performs unit propagation of all unary clauses
    """
    return clauses
    # for clause in clauses:
    #     if len(clause) == 1 and clause.is_hard():
    #         # propagate over all clauses
    #         clauses.remove(clause)
    #         v = clause.pop()
    #         for c in clauses:
    #             if v in c:
    #                 clauses.remove(c)
    #             elif ~v in c:
    #                 c.remove(~v)
    #         # once propagated the value of v, simplify again. until no unitary clauses are left
    #         return simplify_clauses(clauses)
    #
    # # if there are any unitary clause left, just return
    # return clauses

def weight_clauses(clauses, weight):
    """
    for each clause in clauses insert a fresh variable representing whether the specific
    constraint is satisfied or not finally, append a soft constraint imposing that fresh
    variable to be true, assigning the given weight as violation penalty.
    """
    # do nothing if clauses correspond to a hard constraint
    if weight is None or weight is float('inf'):
        return clauses
    else:
        fresh_var = Boolean()
        fresh_var_inv = ~fresh_var
        for clause in clauses:
            clause.add(fresh_var_inv)
        clauses.append(Clause([fresh_var], weight))
        return clauses

class Boolean:

    NUM_VARS = 1

    def __init__(self, num=None):
        if num is not None:
            self.num = num
        else:
            self.num = Boolean.NUM_VARS
            Boolean.NUM_VARS += 1

    def __abs__(self):
        return abs(self.num)

    def __str__(self):
        return str(self.num)

    def __repr__(self):
        return str(self.num)

    def __invert__(self):
        return Boolean(-self.num)

    def __eq__(self, other):
        return self.num == other.num

    def __hash__(self):
        return self.num

class BooleanVec(list):

    def __init__(self, length):
        list.__init__([])
        for i in range(length):
            self.append(Boolean())

class Clause(set):

    TOP = 1

    def __init__(self, args, weight=None):
        set.__init__(set())
        for arg in args:
            self.add(arg)
        if weight is None or weight == float('inf'):
            self.weight = None
        else:
            Clause.TOP += weight
            self.weight = weight

    def is_hard(self):
        return self.weight is None

    def is_soft(self):
        return self.weight is not None

    def is_empty(self):
        return len(self) == 0

    def __str__(self):
        if self.weight is None:
            string = str(Clause.TOP) + ' '
        else:
            string = str(self.weight) + ' '
        for elem in self:
            string += str(elem) + ' '
        return string + '0'

    def __hash__(self):
        import hashlib
        # Assumes the default UTF-8
        hash_object = hashlib.md5((str(self)+str(self.weight)).encode())
        return int(hash_object.hexdigest(), 16)

class CardinalityFactory:

    def __init__(self):
        pass

    def exactly_k(self, args, k):
        return []

    def at_least_k(self, args, k):
        return []

    def at_most_k(self, args, k):
        return []

    def at_most_1(self, args):
        """
        At Most One constraint implemented using ladder encoding
        if len(args) > 14, else the pairwise encoding is used (as it is faster)
        """
        n = len(args)
        phi = []
        if n > 14:
            # constraint all vars in args but both first and the last one
            s = BooleanVec(n-1)
            for i in range(1, n-1):
                phi.append(Clause([s[i], ~s[i-1]]))  # all aux BooleanVec 's' will be of the form [0*011*]
                phi.append(Clause([~args[i], s[i]]))        # args[i] -> s[i]
                phi.append(Clause([~args[i], ~s[i-1]]))     # args[i] -> ~s[i-1]
                # (thus, forcing the only true var to be the one placed at the only 0-1 transition of BooleanVec 's')
            # add special cases for the first and last variable
            phi.extend([Clause([~args[0], s[0]])])
            phi.extend([Clause([~args[n-1], ~s[n-2]])])
        else:
            phi = [Clause([~a, ~b]) for a in args for b in args if abs(a) < abs(b)]

        return phi

class MCardinalityNetwork(CardinalityFactory):
    """
    Implementation described in this paper:
    http://upcommons.upc.edu/e-prints/bitstream/2117/23251/1/Nieuwenhuis.pdf
    """
    def __init__(self, factor=5.0):
        CardinalityFactory.__init__(self)
        self._lambda = factor

    #
    #  Counters
    #

    def count_merge(self, a, b):

        if a < b:
            return self.count_merge(b, a)

        elif a == 1 and b == 1:
            return 2, 3
        elif b == 0:
            return 0, 0
        # len(a) > 1 or len(b) > 1
        else:
            v1, c1 = self.count_merge(a-(a/2), b-(b/2))
            v2, c2 = self.count_merge(a/2, b/2)

            k = (a+b+1)/2
            return v1+v2+2*k, c1+c2+3*k

    def count_sort(self, a):
        if a == 1:
            return 0, 0
        elif a == 2:
            return 2, 3
        else:
            v1, c1 = self.count_sort(a-a/2)
            v2, c2 = self.count_sort(a/2)
            v3, c3 = self.count_merge(a-a/2, a/2)
            return v1+v2+v3, c1+c2+c3

    def count_simplified_merge(self, a, b, c):

        if a < b:
            return self.count_simplified_merge(b, a, c)
        elif a == 1 and b == 1 and c == 1:
            return 1, 2
        elif a > c:
            return self.count_simplified_merge(c, b, c)
        elif b > c:
            return self.count_simplified_merge(a, c, c)
        elif a+b <= c:
            return self.count_merge(a, b)
        else:
            v1, c1 = self.count_simplified_merge(a-a/2, b-b/2, c/2+1)
            v2, c2 = self.count_simplified_merge(a/2, b/2, c/2)
            if c % 2 == 0:
                c3 = (3*c-2)/2+2
            else:
                c3 = (3*c-3)/2
            return v1+v2+c-1, c1+c2+c3

    def count_m_cardinality_network(self, a, m):

        if a <= m:
            return self.count_sort(a)
        else:
            v1, c1 = self.count_m_cardinality_network(a-a/2, m)
            v2, c2 = self.count_m_cardinality_network(a/2, m)
            v3, c3 = self.count_simplified_merge(min(a-a/2, m), min(a/2, m), m)
            return v1+v2+v3, c1+c2+c3

    def count_direct_merge(self, a, b):
        return a+b, a*b+a+b

    def count_direct_sort(self, a):
        return a, (2**a)-1

    def count_direct_simplified_merge(self, a, b, c):

        if a < b:
            return self.count_direct_simplified_merge(b, a, c)
        elif a == 1 and b == 1 and c == 1:
            return 1, 2
        elif a > c:
            return self.count_direct_simplified_merge(c, b, c)
        elif b > c:
            return self.count_direct_simplified_merge(a, c, c)
        elif a+b <= c:
            return self.count_direct_merge(a, b)
        else:
            return c, (a+b)*c - (c*(c-1))/2 - (a*(a-1))/2 - (b*(b-1))/2

    def count_direct_m_cardinality_network(self, a, m):
        if a <= m:
            return self.count_direct_sort(a)
        else:
            clauses = 0
            n_fact = math.factorial(a)
            for i in range(1, m+1):
                clauses += n_fact/(math.factorial(i)*math.factorial(a-i))
            return m, clauses

    #
    # Encodings
    #

    def two_comparator(self, a, b, at_most=True, at_least=False):
        c1 = Boolean()
        c2 = Boolean()
        clauses = []
        if at_most:
            clauses.append(Clause([~a, c1]))        # a -> c1
            clauses.append(Clause([~b, c1]))        # b -> c1
            clauses.append(Clause([~a, ~b, c2]))    # a^b -> c2
        if at_least:
            clauses.append(Clause([~c1, a, b]))     # c1 -> avb
            clauses.append(Clause([~c2, a]))        # c2 -> a
            clauses.append(Clause([~c2, b]))        # c2 -> b
        return [c1, c2], clauses

    def merge(self, a_args, b_args, at_most=True, at_least=False):

        ka = len(a_args)
        kb = len(b_args)

        if ka == 1 and kb == 1:
            return self.two_comparator(a_args[0], b_args[0], at_most, at_least)

        elif ka == 0:
            return b_args, []

        elif ka % 2 == 0 and kb % 2 == 0 and kb > 0:
            odds, phi1 = self.mixed_merge([a_args[i] for i in range(0, ka, 2)], [b_args[i] for i in range(0, kb, 2)], at_most, at_least)
            even, phi2 = self.mixed_merge([a_args[i] for i in range(1, ka, 2)], [b_args[i] for i in range(1, kb, 2)], at_most, at_least)
            phi = phi1 + phi2
            results = []
            for i in range(len(even)-1):
                ys, cls = self.two_comparator(even[i], odds[i+1], at_most, at_least)
                results.extend(ys)
                phi.extend(cls)
            return [odds[0]] + results + [even[-1]], phi

        elif ka % 2 == 0 and kb > 0:
            odds, phi1 = self.mixed_merge([a_args[i] for i in range(0, ka, 2)], [b_args[i] for i in range(0, kb, 2)], at_most, at_least)
            even, phi2 = self.mixed_merge([a_args[i] for i in range(1, ka, 2)], [b_args[i] for i in range(1, kb, 2)], at_most, at_least)
            phi = phi1 + phi2
            results = []
            for i in range(len(even)):
                ys, cls = self.two_comparator(even[i], odds[i+1], at_most, at_least)
                results.extend(ys)
                phi.extend(cls)
            return [odds[0]] + results, phi

        elif ka % 2 == 1 and kb % 2 == 1 and kb > 0:
            odds, phi1 = self.mixed_merge([a_args[i] for i in range(0, ka, 2)], [b_args[i] for i in range(0, kb, 2)], at_most, at_least)
            even, phi2 = self.mixed_merge([a_args[i] for i in range(1, ka, 2)], [b_args[i] for i in range(1, kb, 2)], at_most, at_least)
            phi = phi1 + phi2
            results = []
            for i in range(len(even)):
                ys, cls = self.two_comparator(even[i], odds[i+1], at_most, at_least)
                results.extend(ys)
                phi.extend(cls)
            return [odds[0]] + results + [odds[-1]], phi

        else:
            return self.merge(b_args, a_args, at_most, at_least)

    def simplified_merge(self, a_args, b_args, c, at_most=True, at_least=False):

        ka = len(a_args)
        kb = len(b_args)

        if ka == 1 and kb == 1 and c == 1:
            y = Boolean()
            clauses = []
            if at_most:
                clauses.extend([Clause([~a_args[0], y]), Clause([~b_args[0], y])])
            if at_least:
                clauses.append(Clause([~y, a_args[0], b_args[0]]))
            return [y], clauses

        elif ka > c:
            return self.mixed_simplified_merge(a_args[:c-ka], b_args, at_most, at_least)

        elif kb > c:
            return self.mixed_simplified_merge(a_args, b_args[:c-kb], at_most, at_least)

        elif ka+kb <= c:
            return self.mixed_merge(a_args, b_args, at_most, at_least)

        # ka,kb <= c ka+kb > c
        elif c % 2 == 0:

            odds, phi1 = self.mixed_simplified_merge([a_args[i] for i in range(0, ka, 2)],
                                         [b_args[i] for i in range(0, kb, 2)], c/2+1, at_most, at_least)
            even, phi2 = self.mixed_simplified_merge([a_args[i] for i in range(1, ka, 2)],
                                               [b_args[i] for i in range(1, kb, 2)], c/2, at_most, at_least)
            phi = phi1 + phi2
            results = []
            for i in range(len(even)-1):
                ys, cls = self.two_comparator(even[i], odds[i+1], at_most, at_least)
                results.extend(ys)
                phi.extend(cls)

            yc = Boolean()
            if at_most:
                phi.extend([Clause([~even[-1], yc]), Clause([~odds[-1], yc])])
            if at_least:
                phi.append(Clause([~yc, even[-1], odds[-1]]))

            return [odds[0]] + results + [yc], phi

        # c is odd
        else:
            odds, phi1 = self.mixed_simplified_merge([a_args[i] for i in range(0, ka, 2)],
                                         [b_args[i] for i in range(0, kb, 2)], (c+1)/2, at_most, at_least)
            even, phi2 = self.mixed_simplified_merge([a_args[i] for i in range(1, ka, 2)],
                                               [b_args[i] for i in range(1, kb, 2)], (c-1)/2, at_most, at_least)
            phi = phi1 + phi2
            results = []
            for i in range(len(even)):
                ys, cls = self.two_comparator(even[i], odds[i+1], at_most, at_least)
                results.extend(ys)
                phi.extend(cls)

            return [odds[0]] + results, phi

    def sort(self, args, at_most=True, at_least=False):
        n = len(args)

        if n == 1:
            return args, []

        elif n == 2:
            return self.mixed_merge([args[0]], [args[1]], at_most, at_least)

        elif n > 2:
            left_hand, phi1 = self.mixed_sort(args[:n/2], at_most, at_least)
            right_hand, phi2 = self.mixed_sort(args[n/2:], at_most, at_least)
            result, phi3 = self.mixed_merge(left_hand, right_hand, at_most, at_least)
            return result, phi1 + phi2 + phi3

    def m_cardinality_network(self, args, m, at_most=True, at_least=False):
        n = len(args)

        if n <= m:
            return self.mixed_sort(args, at_most, at_least)
        else:
            l_side, phi1 = self.mixed_m_cardinality_network(args[:n/2], m, at_most, at_least)
            r_side, phi2 = self.mixed_m_cardinality_network(args[n/2:], m, at_most, at_least)
            result, phi3 = self.mixed_simplified_merge(l_side, r_side, m, at_most, at_least)
            return result, phi1 + phi2 + phi3

    def direct_merge(self, a_args, b_args):
        result = BooleanVec(len(a_args)+len(b_args))
        phi = []
        for i in range(len(a_args)):
            phi.append(Clause([~a_args[i], result[i]]))

            for j in range(len(b_args)):
                phi.append(Clause([~a_args[i], ~b_args[j], result[i+j+1]]))

        for j in range(len(b_args)):
            phi.append(Clause([~b_args[j], result[j]]))

        return result, phi

    def direct_sort(self, args):

        n = len(args)
        phi = []
        result = []

        for k in range(1, n+1):
            yk = Boolean()
            combs = [list(combination) for combination in itertools.combinations(range(n), k)]
            for combination in combs:
                clause = Clause([~args[i] for i in combination])
                clause.add(yk)
                phi.append(clause)
            result.append(yk)

        return result, phi

    def direct_simplified_merge(self, a_args, b_args, c):

        ka = len(a_args)
        kb = len(b_args)

        if ka == 1 and kb == 1 and c == 1:
            y = Boolean()
            clauses = [Clause([~a_args[0], y]), Clause([~b_args[0], y])]
            return [y], clauses

        elif ka > c:
            return self.direct_simplified_merge(a_args[:c-ka], b_args)

        elif kb > c:
            return self.direct_simplified_merge(a_args, b_args[:c-kb])

        elif ka+kb <= c:
            return self.direct_merge(a_args, b_args)

        # ka,kb <= c ka+kb > c
        else:
            result = BooleanVec(c)
            phi = []
            for i in range(len(a_args)):
                phi.append(Clause([~a_args[i], result[i]]))

                for j in range(len(b_args)):
                    if i+j+1 < c:
                        phi.append(Clause([~a_args[i], ~b_args[j], result[i+j+1]]))

            for j in range(len(b_args)):
                phi.append(Clause([~b_args[j], result[j]]))

            return result, phi

    def direct_m_cardinality_network(self, args, m):

        n = len(args)

        if n <= m:
            return self.direct_sort(args)
        else:
            phi = []
            result = []

            for k in range(1, m+1):
                yk = Boolean()
                combs = [list(combination) for combination in itertools.combinations(range(n), k)]
                for combination in combs:
                    clause = Clause([~args[i] for i in combination])
                    clause.add(yk)
                    phi.append(clause)
                result.append(yk)

            return result, phi

    def mixed_merge(self, a_args, b_args, at_most=True, at_least=False):

        if at_least:
            return self.merge(a_args, b_args, at_most, at_least)

        vr, cr = self.count_merge(len(a_args), len(b_args))
        vd, cd = self.count_direct_merge(len(a_args), len(b_args))

        recursive = self._lambda*vr+cr
        direct = self._lambda*vd+cd
        if recursive <= direct:
            return self.merge(a_args, b_args)
        else:
            return self.direct_merge(a_args, b_args)

    def mixed_sort(self, args, at_most=True, at_least=False):

        if at_least:
            return self.sort(args, at_most, at_least)

        vr, cr = self.count_sort(len(args))
        vd, cd = self.count_direct_sort(len(args))

        recursive = self._lambda*vr+cr
        direct = self._lambda*vd+cd
        if recursive <= direct:
            return self.sort(args)
        else:
            return self.direct_sort(args)

    def mixed_simplified_merge(self, a_args, b_args, c, at_most=True, at_least=False):

        if at_least:
            return self.simplified_merge(a_args, b_args, c, at_most, at_least)

        vr, cr = self.count_simplified_merge(len(a_args), len(b_args), c)
        vd, cd = self.count_direct_simplified_merge(len(a_args), len(b_args), c)

        recursive = self._lambda*vr+cr
        direct = self._lambda*vd+cd
        if recursive <= direct:
            return self.simplified_merge(a_args, b_args, c)
        else:
            return self.direct_simplified_merge(a_args, b_args, c)

    def mixed_m_cardinality_network(self, args, m, at_most=True, at_least=False):

        if at_least:
            return self.m_cardinality_network(args, m, at_most, at_least)

        vr, cr = self.count_m_cardinality_network(len(args), m)
        vd, cd = self.count_direct_m_cardinality_network(len(args), m)

        recursive = self._lambda*vr+cr
        direct = self._lambda*vd+cd
        if recursive <= direct:
            return self.m_cardinality_network(args, m)
        else:
            return self.direct_m_cardinality_network(args, m)


    def weighted_card(self, args, lb, minimum, ub, maximum, weight):
        """
        Introduces weighted cardinality constraints acording with XHSTT cost format
        :param args: list with implied variables
        :param lb:  lower bound that the variables provided are guaranteed to achieve
        :param min: minimum cardinality to constraint
        :param ub: upper bound that the variables provided are guaranteed to achieve
        :param max: maximum cardinality to constraint
        :param weight: weight for breaking in ONE step any limit (either min or max)
        """

        #ub = min(ub, maximum+1)

        if lb < 0:
            raise Exception('negative Lower Bound!!')

        if weight == float('inf'):

            phi = []

            if lb < minimum < maximum:
                phi = self.at_least_k(args, minimum)
            if minimum < maximum < ub:
                phi.extend(self.at_most_k(args, maximum))
            if minimum == maximum:
                phi = self.exactly_k(args, minimum)

            return phi

        elif minimum == maximum == 0:
            return [Clause([~var], weight) for var in args]

        elif minimum > lb and maximum < ub:

            q, phi = self.m_cardinality_network(args, ub, True, True)

            # lower end:
            for i in range(0, lb):
                phi.append(Clause([q[i]]))
            for i in range(lb+1, minimum+1):
                if i > 0:
                    phi.append(Clause([q[i-1]], weight))

            # high end:
            for i in range(maximum, ub):
                phi.append(Clause([~q[i]], weight))

            return phi

        elif minimum == lb:
            if maximum < ub:
                q, phi = self.m_cardinality_network(args, ub, True, False)
                # lower end:
                for i in range(maximum-1):
                    phi.append(Clause([q[i]]))
                # high end:
                for i in range(maximum, ub):
                    phi.append(Clause([~q[i]], weight))

                return phi
            return []
        # max == ub
        else:
            q, phi = self.m_cardinality_network(args, ub, False, True)

            # lower end:
            for i in range(0, lb-1):
                phi.append(Clause([q[i]]))
            for i in range(lb, minimum+1):
                if i > 0:
                    phi.append(Clause([q[i-1]], weight))
            # high end:
            for i in range(minimum+1, ub):
                phi.append(Clause([~q[i]]))

            return phi


    def exactly_k(self, args, k):
        # return self.at_least_k(args, k) + self.at_most_k(args, k)
        if k < 0 or k > args:
            raise Exception('Error: exactly k with wrong arguments')
        if k == 0:
            return [Clause([~v]) for v in args]
        if k == 1:
            return self.at_least_k(args, 1) + self.at_most_k(args, 1)
        else:
            q, phi = self.m_cardinality_network(args, k+1, True, True)

            # Impose all result variables above k-th to be false
            phi.append(Clause([~q[k]]))

            # Impose all first-k result variables to be true
            # (since it's not complete, it won't affect the input and it will reduce search space by fixing out vars)
            for i in range(k):
                phi.append(Clause([q[i]]))

            return simplify_clauses(phi)

    def at_least_k(self, args, k):
        if k == 0:
            return []
        elif k == 1:
            return [Clause(args)]
        else:
            # return self.at_most_k([~v for v in args], len(args)-k)
            q, phi = self.m_cardinality_network(args, k, False, True)

            # Impose all first-k result variables to be true
            # (since it's not complete, it won't affect the input and it will reduce search space by fixing out vars)
            for i in range(k):
                phi.append(Clause([q[i]]))

            return simplify_clauses(phi)

    def at_most_k(self, args, k):

        if k >= len(args):
            return []
        if k == 0:
            return [Clause([~var]) for var in args]
        if k == 1:
            return self.at_most_1(args)
        else:
            result, phi = self.mixed_m_cardinality_network(args, k+1)

            # Impose all result variables above k-th to be false
            phi.append(Clause([~result[k]]))

            # Impose all first-k result variables to be true
            # (since it's not complete, it won't affect the input and it will reduce search space by fixing out vars)
            for i in range(k):
                phi.append(Clause([result[i]]))

            return phi

class ModuloTotalizer(CardinalityFactory):

    def __init__(self):
        CardinalityFactory.__init__(self)
        self.sort_network = MCardinalityNetwork(5)


    def modulo_unary_adder(self, f, a, g, b, p):
        """
        :param f: (list) quotient of the lhs operator (lhs = fxp+a)
        :param a: (list) remainder of the lhs operator                  pre: len(a) < p
        :param g: (list) quotient of the rhs operator (rhs = gxp+a)
        :param b: (list) remainder of the rhs operator                  pre: len(b) < p
        :param p: (int)  modulo such that n = quotient x p + remainder
        :return: h, r, phi. where h stands for the quotient, r for the remainder and phi for the encoded clauses
        """
        r = BooleanVec(p-1)
        c = Boolean()  # carry

        phi = [Clause([~a[i], r[i], c]) for i in range(len(a))] + [Clause([~b[i], r[i], c]) for i in range(len(b))]
        for i in range(len(a)):
            for j in range(len(b)):
                if i+j < p-2:
                    phi.append(Clause([~a[i], ~b[j], r[i+j+1], c]))
                elif i+j > p-2:
                    phi.append(Clause([~a[i], ~b[j], r[(i+j+1) % p]]))
                    phi.append(Clause([~a[i], ~b[j], c]))
                else:
                    phi.append(Clause([~a[i], ~b[j], c]))

        h = BooleanVec(len(f)+len(g)+1)
        phi.append(Clause([~c, h[0]]))
        phi.extend([Clause([~f[i], h[i]]) for i in range(len(f))] + [Clause([~g[i], h[i]]) for i in range(len(g))])
        phi.extend([Clause([~f[i], ~c, h[i+1]]) for i in range(len(f))])
        phi.extend([Clause([~g[i], ~c, h[i+1]]) for i in range(len(g))])
        for i in range(len(f)):
            for j in range(len(g)):
                phi.append(Clause([~f[i], ~g[j], h[i+j+1]]))
                phi.append(Clause([~f[i], ~g[j], ~c, h[i+j+2]]))

        return h, r, phi

    def unary_adder(self, a, b):
        """
        :param a: (list) operands of lhs operator
        :param b: (list) operands of rhs operator
        :return: (list, list) bits of the result plus a list of the encoding clauses
        """
        r = BooleanVec(len(a)+len(b))

        phi = [Clause([~a[i], r[i]]) for i in range(len(a))] + [Clause([~b[i], r[i]]) for i in range(len(b))]
        for i in range(len(a)):
            for j in range(len(b)):
                phi.append(Clause([~a[i], ~b[j], r[i+j+1]]))

        return r, phi

    def totalizer(self, seq):
        n = len(seq)
        if n == 1:
            return seq, []
        else:
            a, phi1 = self.totalizer(seq[:n/2])
            b, phi2 = self.totalizer(seq[n/2:])
            o, phi3 = self.unary_adder(a, b)
            return o, phi1 + phi2 + phi3


    def modulo_totalizer(self, seq, p):
        n = len(seq)
        if n < p:
            r, phi = self.totalizer(seq)
            r, phi = self.sort_network.direct_sort(seq)
            return [], r, phi
        else:
            f, a, phi1 = self.modulo_totalizer(seq[:n/2], p)
            g, b, phi2 = self.modulo_totalizer(seq[n/2:], p)
            h, r, phi3 = self.modulo_unary_adder(f, a, g, b, p)
            return h, r, phi1 + phi2 + phi3

    def at_most_k(self, seq, k):

        if k == 0:
            return [Clause([~v]) for v in seq]
        if k == 1:
            return self.at_most_1(seq)

        n = len(seq)
        p = max(int(math.sqrt(n)), 7)
        # k modulo p = sigma x p + remainder
        sigma = k / p
        remainder = k % p
        # get the vars expression in modular form
        h, r, phi = self.modulo_totalizer(seq, p)

        # Encoded as 'The comparator' in Modulo Based CNF Encoding of Cardinality Constraints and Its Applications paper
        # Encoding card(seq) <= k
        for i in range(sigma, len(h)):
            phi.append(Clause([~h[i]]))
        for j in range(remainder, len(r)):
            if sigma > 0:
                phi.append(Clause([~h[sigma-1], ~r[j]]))
            else:
                phi.append(Clause([~r[j]]))

        return phi

    def at_least_k(self, seq, k):
        if k == 0:
            return []
        elif k == 1:
            return [Clause(seq)]
        else:
            return self.at_most_k([~v for v in seq], len(seq)-k)

    def exactly_k(self, seq, k):
        if k == 0:
            return [Clause([~var]) for var in seq]
        if k == 1:
            return self.at_most_1(seq) + [Clause(seq)]
        phi1 = self.at_most_k(seq, k)
        phi2 = self.at_least_k(seq, k)
        return phi1 + phi2

    def weighted_card(self, args, lb, minimum, ub, maximum, weight):
        """
        Introduces weighted cardinality constraints acording with XHSTT cost format
        :param args: list with implied variables
        :param lb:  lower bound that the variables provided are guaranteed to achieve
        :param min: minimum cardinality to constraint
        :param ub: upper bound that the variables provided are guaranteed to achieve
        :param max: maximum cardinality to constraint
        :param weight: weight for breaking in ONE step any limit (either min or max)
        """

        if lb < 0:
            raise Exception('negative Lower Bound!!')

        phi = []

        if weight == float('inf'):

            if lb < minimum < maximum:
                phi = self.at_least_k(args, minimum)
            if minimum < maximum < ub:
                phi.extend(self.at_most_k(args, maximum))
            if minimum == maximum:
                phi = self.exactly_k(args, minimum)

            return phi

        elif minimum == maximum == 0:
            return [Clause([~var], weight) for var in args]

        # encode minimum
        if minimum > lb:
            phi = self.weighted_at_most_k([~arg for arg in args], len(args)-minimum, len(args)-lb, weight)
        # encode maximum
        if maximum < ub:
            phi.extend(self.weighted_at_most_k(args, maximum, ub, weight))

        return phi

    def weighted_at_most_k(self, args, k, ub, weight):

        n = len(args)
        p = max(int(math.sqrt(n)), 7)

        # get the vars expression in modular form
        h, r, phi = self.modulo_totalizer(args, p)

        # Encoding card(seq) <= k
        # Encoding card(seq) <= k
        for y in range(k, ub):

            sigma = y / p
            remainder = y % p
            aux = Boolean()  # protection variable

            if h:
                phi.append(Clause([~aux, ~h[sigma]]))

            for j in range(remainder, len(r)):
                if sigma > 0 and h:
                    phi.append(Clause([~aux, ~h[sigma-1], ~r[j]]))
                else:
                    phi.append(Clause([~aux, ~r[j]]))

            phi.append(Clause([aux], weight))

        return phi
