"""
 Copyright 2015 Cristofor Nogueira Gascons
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = 'cris nogueira'

from xmlparser import *
from collections import OrderedDict


# Global counters for ModelEntities 'num' field
_tim_c = 0  # Times counter
_eve_c = 0  # Events counter
_res_c = 0  # Resources counter
_rty_c = 0  # ResourceType counter

class ModelXHSTT:

    def __init__(self):
        self.events = OrderedDict()
        self.resources = OrderedDict()
        self.times = OrderedDict()

        self.time_groups = dict()
        self.event_groups = dict()
        self.resource_groups = dict()

        self.rtypes = dict()

        # declare all 'num2' dictionaries (allowing fast fetching of elements by its elements.num attribute)
        self.num2event = dict()
        self.num2resource = dict()
        self.num2time = dict()
        self.num2rtype = dict()

    def __str__(self):
        return str(self.times) + '\n' + str(self.events) + '\n' + str(self.resources)

    def register_time(self, identifier, name):
        if identifier not in self.times:
            time = Time(identifier, name)
            self.times[identifier] = time
            self.num2time[time.num] = time
        else:
            raise Exception('Multiple Times with the same id')

    def register_event(self, identifier, name, duration, color=None):
        if identifier not in self.events:
            event = Event(identifier, name, duration, color)
            self.events[identifier] = event
            self.num2event[event.num] = event
        else:
            raise Exception('Multiple Events with the same id')

    def register_resource(self, identifier, name, rtype):
        if rtype in self.rtypes:
            if identifier not in self.resources:
                resource = Resource(identifier, name, rtype)
                self.resources[identifier] = resource
                self.num2resource[resource.num] = resource
                self.rtypes[rtype].add_resource(resource)
            else:
                raise Exception('Multiple Resources with the same id')
        else:
            raise Exception('Attempt to register a resource with an unregistered resource type')

    def register_resource_type(self, identifier, name):
        if identifier not in self.rtypes:
            rtype = ResourceType(identifier, name)
            self.rtypes[identifier] = rtype
            self.num2rtype[rtype.num] = rtype
        else:
            raise Exception('Multiple declaration of same resource type')

    def get_time_by_ref(self, reference):
        if reference in self.times:
            return self.times[reference]
        else:
            return None

    def get_resource_by_ref(self, reference):
        if reference in self.resources:
            return self.resources[reference]
        else:
            return None

    def get_event_by_ref(self, reference):
        if reference in self.events:
            return self.events[reference]
        else:
            return None

    def declare_time_group(self, identifier, name, tag=None):
        if identifier not in self.time_groups:
            self.time_groups[identifier] = Group(identifier, name, 'TimeGroup', tag)
        else:
            raise Exception('Already defined TimeGroup')

    def declare_resource_group(self, identifier, name, rtype):
        if rtype in self.rtypes:
            if identifier not in self.resource_groups:
                self.resource_groups[identifier] = Group(identifier, name, 'ResourceGroup', rtype)
            else:
                raise Exception('Already defined ResourceGroup')
        else:
            raise Exception('Undeclared resource type')

    def declare_event_group(self, identifier, name, tag=None):
        if identifier not in self.event_groups:
            self.event_groups[identifier] = Group(identifier, name, 'EventGroup', tag)
        else:
            raise Exception('Already defined EventGroup')

    def time_to_group(self, time_id, group_ref):

        if time_id not in self.times or group_ref not in self.time_groups:
            raise Exception('Undeclared reference encountered')
        else:
            # Add group reference to the time's groups
            self.times[time_id].groups.add(group_ref)

            # Register Time to the group member's
            self.time_groups[group_ref].elems.add(time_id)

    def resource_to_group(self, res_id, group_ref):

        if res_id not in self.resources or group_ref not in self.resource_groups:
            raise Exception('Undeclared reference encountered')
        elif self.resources[res_id].rtype != self.resource_groups[group_ref].opt:
            raise Exception('Attached resource to a group of different type')
        else:
            # Add group reference to the resource's groups
            self.resources[res_id].groups.add(group_ref)

            # Register Resource to the group member's
            self.resource_groups[group_ref].elems.add(res_id)

    def event_to_group(self, event_id, group_ref):

        if event_id not in self.events or group_ref not in self.event_groups:
            raise Exception('Undeclared reference encountered')
        else:
            # Add group reference to the event's groups
            self.events[event_id].groups.add(group_ref)

            # Register Event to the group member's
            self.event_groups[group_ref].elems.add(event_id)

    def upper_bound_by_resource_type(self, num):
        """
        Function used to determine the upper bound of a variable representing a given resource type
        :param num: Number that is bound to a resource type
        :return: number of resources in self.resources that have its type bounded to 'num'
        """
        return self.num2rtype[num].ub

    def get_events_from_group(self, group_ref):

        if group_ref is None:
            raise Exception('empty group_ref')

        return frozenset(self.event_groups[group_ref].elems)

    def get_resources_from_group(self, group_ref):

        if group_ref is None:
            raise Exception('empty group_ref')

        return frozenset(self.resource_groups[group_ref].elems)

    def get_times_from_group(self, group_ref):

        if group_ref is None:
            raise Exception('empty group_ref')

        return frozenset(self.time_groups[group_ref].elems)

    def set_time_for_event(self, event_num, duration, start_t):
        pass

    def preferResourcesConstraint(self, events_ids, resources_ids, role):
        pass

    def preferTimesConstraint(self, cost, events_ids, times_ids, duration):
        pass

    def avoidClashesConstraint(self, cost, resources_ids):
        pass

    def splitEventsConstraint(self, cost, events, minimum, maximum, min_amount, max_amount):
        pass

    def spreadEventsConstraint(self, cost, event_groups, time_groups):
        """
        time_groups is a dictionary of tuples. where keys are time_group references
        and values are tuples of the kind (minimum, maximum)
        """
        pass

    def avoidUnavailableTimesConstraint(self, cost, resource_ids, time_ids):
        pass

    def distributeSplitEventsConstraints(self, cost, events_ids, duration, minimum, maximum):
        pass

    def limitIdleTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):
        pass

    def clusterBusyTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):
        pass

    def onParsedEvents(self, spread_constraint=True):
        pass

    def get_times_assignments(self):
        pass

class Group:

    def __init__(self, identifier, name, gtype, opt=None):
        self.identifier = identifier
        self.name = name
        self.gtype = gtype  # group type: {TimeGroup, EventGroup, ResourceGroup}
        self.opt = opt      # optional info about group (whether it represents a Day, Week, Course..)
        self.elems = set()  # identifiers of elements forming this group

    def __str__(self):
        if self.opt is not None:
            return self.name + ' (' + self.opt + ')'
        else:
            return self.name + ' (' + self.gtype + ')'

class ModelEntity:

    def __init__(self, identifier, name='NAMELESS'):
        self.identifier = identifier
        self.name = name
        self.num = -1
        self.groups = set()

    def __str__(self):
        return self.num

class ResourceType(ModelEntity):

    def __init__(self, identifier, name='NAMELESS_RESOURCE_TYPE'):
        ModelEntity.__init__(self, identifier, name)
        self.type = 'resource_type'
        self.lb = 0  # lower bound
        self.ub = 0  # upper bound
        self.domain = set()

        # Assign to the object a unique identifier among other instances
        global _rty_c
        self.num = _rty_c
        _rty_c += 1

    def add_resource(self, resource):
        self.groups.add(resource)
        self.domain.add(resource.num)

        # update upper and lower bounds
        if resource.num > self.ub:
            self.ub = resource.num
        elif resource.num < self.lb:
            self.lb = resource.num

class Time(ModelEntity):

    def __init__(self, identifier, name='NAMELESS_TIME'):
        ModelEntity.__init__(self, identifier, name)
        self.type = 'time'

        # Assign to the object a unique identifier among other instances
        global _tim_c
        self.num = _tim_c
        _tim_c += 1

class Resource(ModelEntity):

    def __init__(self, identifier, name='NAMELESS_RESOURCE', resource_type=None):
        ModelEntity.__init__(self, identifier, name)

        if resource_type is None:
            raise Exception('Cannot create a Resource without type')

        self.type = 'resource'
        self.rtype = resource_type

        # Assign to the object a unique identifier among other instances
        global _res_c
        self.num = _res_c
        _res_c += 1

class Event(ModelEntity):

    def __init__(self, identifier, name='NAMELESS_EVENT', duration=1, color=None):
        ModelEntity.__init__(self, identifier, name)

        # Assign to the object a unique identifier among other instances
        global _eve_c
        self.num = _eve_c
        _eve_c += 1

        self.type = 'event'
        self.color = color
        self.duration = duration
        self.time = None
        self.resources = dict()
        self.mapping = dict()
        self.needed = set()  # set of needed ResourceType's nums

    def hasRole(self, role):
        return role in self.resources

    def hasPreassignedResource(self, role):
        return self.resources[role] is not None

    def getPreassignedResources(self):
        """
        Returns a frozen set containing all non None resources attached to the event
        """
        return frozenset([r for r in self.resources.itervalues() if r is not None])



    def getPreassignedResource(self, role):
        return self.resources[role]

    def hasPreassignedTime(self):
        return self.time is not None

    def roles(self):
        return self.resources.keys()

    def needed_resources(self):
        return self.needed

    def is_preassigned(self, num):
        if self.get_preassigned(num) is None:
            return False
        else:
            return True

    def get_preassigned(self, num):
        for resource in self.resources.itervalues():
            if resource is not None and resource.num == num:
                return resource
        return None

    def get_preassigned_resources_nums(self):
        """
        returns a set containing all numerals of preassigned resources
        """
        preassigned_nums = set()
        for resource in self.resources.itervalues():
            if resource is not None:
                preassigned_nums.add(resource.num)
        return preassigned_nums

    def set_time(self, time_ref):
        self.time = time_ref

    def attach_resource(self, resource, role, resource_type, model):

        if resource is None:

            if role is None or resource_type is None:
                raise Exception('Cannot attach a resource without role nor resource_type')

            self.resources[role] = None
            self.mapping[role] = resource_type

            num = model.rtypes[resource_type].num

        else:

            # if the resource has given resource_type assert its correctness
            if resource_type is None or resource.rtype == resource_type:

                if role is None:

                    # Find an unused role name for the resource
                    i = 0
                    role = 'role_' + str(i)
                    while role in self.resources:
                        i += 1
                        role = 'role_' + str(i)

                    self.resources[role] = resource
                    if resource_type is None:
                        self.mapping[role] = resource.rtype
                    else:
                        self.mapping[role] = resource_type

                else:
                    self.resources[role] = resource
                    if resource_type is None:
                        self.mapping[role] = resource.rtype
                    else:
                        self.mapping[role] = resource_type

            else:
                raise Exception('Malformed input given a wrong resource_type for the preassigned resource')

            num = model.rtypes[resource.rtype].num

        if num in self.needed:
            print 'Warning! Encountered an Event which needs two or more resources of the same type'
        else:
            self.needed.add(num)
