"""
 Copyright 2015 Cristofor Nogueira Gascons
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = 'cris nogueira'

from model import ModelXHSTT
from xmlparser import Parser
from z3 import *
from optimizer import Optimizer
import time
import xmlWritter

_unique_id = 0

class ModelBV(ModelXHSTT):

    def __init__(self, xml_source, timeout=3600000):
        ModelXHSTT.__init__(self)
        
        # Set default optimizer
        self.optimizer = Optimizer()
        
        # Variable declaration
        self.Xt = dict()            # event times
        self.Xs = dict()            # event starting times
        self.Xd = dict()            # event durations
        self.Xr = dict()            # dictionary containing, for each resource all the events that demand such resource

        Parser(self, xml_source).parse_model()

        print 'parsing stage complete'


        print 'all set, starting check'
        tic = time.time()
        self.model = self.optimizer.binary_optimization(timeout)  # 3 hours in milliseconds

        if self.model is not None:
            print 'sat :)'
            x = xmlWritter.XMLWritter(xml_source, self)
            del self.model

        print 'took ', time.time()-tic, ' seconds'


    def get_times_assignments(self):
        time_count = len(self.times)
        res = []
        if self.model is not None:
            for event in self.events.itervalues():
                for d in range(1, event.duration+1):
                    xd_as_str = self.model[self.Xd[event.num][d]].sexpr()[2:]

                    # for each bit set, create an entry
                    for i in range(0, time_count):
                        if xd_as_str[i] == '1':
                            res.append((event.identifier, d, self.num2time[i].identifier))


        return res

    def onParsedEvents(self,spread_constraint=True):
        """
        For each parsed event, declare all related variables. that is time and resource variables
        alongside its known domains (albeit some restrictions still to be known such as prefer-like constraints)

        Note that on this stage, all event object contains only the needed declarations. That is in the extreme case
        that any event would have any role declared without any resource referenced nor assign_resource constraint,
        such role would only appear in the xml and not in it's object representation. Thus all events contains only roles
        for which either have a preassigned resource or solver needs to assign one to it.
        """
        Xt = self.Xt
        Xs = self.Xs
        Xd = self.Xd

        # initialize Xr
        Xr = self.Xr
        for resource in self.resources.itervalues():
            Xr[resource.num] = set()

        time_count = len(self.times)
        counter = 0
        sys.stdout.write('Parsing events ...')
        constraints = []
        for event in self.events.itervalues():
            counter += 1
            sys.stdout.write('\rParsing events\t\t%i/%i' % (counter, len(self.events)))
            e = event.num

            # time variables
            Xt[e] = BitVec('Xt_%i' % e, time_count)
            Xs[e] = BitVec('Xs_%i' % e, time_count)

            # preassigned times
            if event.hasPreassignedTime():
                if event.duration != 1:
                    raise Exception('Encountered an Event with preassigned time with a duration greater than 1')

                preassigned = self.times[event.time].num
                constraints.append(And(Xt[e] == Const_BV(time_count, [preassigned]), Xs[e] == Xt[e]))
            else:
                constraints.append(exactly_k_bits(Xt[e], event.duration))
                constraints.append(Xs[e] != BitVecVal(0, time_count))
                # Xs is contained within Xt
                constraints.append(Xs[e] | Xt[e] == Xt[e])
                # If Event is tanking place in Xt and not in Xt-1 then it must be a starting time
                msb = Const_BV(time_count, [0])  # BV with only the most significant bit set
                constraints.append(Xs[e] == ((LShR(~Xt[e], 1) | msb) & Xt[e]) | Xs[e])

            # Xd constraints:
            #   i -> stands for the BitVector representing the starting times with duration 'i'
            Xd[e] = dict()
            for i in range(1, event.duration+1):
                Xd[e][i] = BitVec('Xd_%i_%i' % (e, i), time_count)
                # If a subevent of duration d has been assigned and immediately after,
                # the event is still taking place, then assign that time as starting time
                constraints.append(Xs[e] == Xs[e] | LShR((Xt[e] << i) & Xd[e][i], i))
                # When a bit of Xd[e][i] is set, ensure that the event takes place in 'i' consecutive slots
                constraints.append(Xd[e][i] == Xd[e][i] & bv_and([(Xt[e] << d) for d in range(0, i)]))

            if event.duration > 1:
                for i in range(1, event.duration+1):
                    # If a duration is set for a specific time, make sure no other durations are set on that time
                    K = BitVec('K#_%i_%i' % (e, _get_unique_id()), time_count)
                    k_args = [(~Xd[e][d]) for d in range(1, event.duration+1) if d != i]
                    constraints.append(K == bv_and(k_args))
                    constraints.append(K == K | Xd[e][i])
                    # If an event with duration d is taking place, make sure no other events start while
                    # the event is taking place (i.e. within the duration range)
                    if i > 1:
                        constraints.append(Xs[e] == Xs[e] & bv_and([~LShR(Xd[e][i], ii) for ii in range(1, i)]))

                    # If a session of duration d starts at timeslot i, then make sure that the event actually
                    # takes place in the next d-1 timeslots and no other event starts at those times.
                    # constraints.append(Xd[e][i] == bv_and([(Xt[e] << d) for d in range(0, i)] +
                    #                                       [(~Xs[e] << d) for d in range(1, i)] +
                    #                                       [(Xs[e] << i | ~Xt[e] << i)]))
            # Avoid clashes constraint for Xd
            for d1 in range(1, event.duration):
                for d2 in range(d1+1, event.duration+1):
                    constraints.append(Xd[e][d1] & Xd[e][d2] == BitVecVal(0, time_count))


            # Impose Xs to be the Or of all durations
            constraints.append(Xs[e] == bv_or([Xd[e][d] for d in range(1, event.duration+1)]))

            # resource variables (this model requires all resources to be preassigned)
            # check that the event has all resources preassigned
            preassigned_resources = event.get_preassigned_resources_nums()
            if len(preassigned_resources) < len(event.resources):
                raise Exception('At least one event with one resource left to assign')

            # Add event reference into all preassigned resources entries in Xr
            for res_num in preassigned_resources:
                Xr[res_num].add(e)

        sys.stdout.write('\n')
        self.optimizer.assert_expression(And(constraints))  # assert expressions as hard-constraints

    def avoidClashesConstraint(self, cost, resources_ids):

        i = 0
        sys.stdout.write('Asserting Avoid Clashes\t\t%i/%i' % (i, len(resources_ids)))
        weight = cost/len(resources_ids)

        zero = BitVecVal(0, len(self.times))
        for resource_id in resources_ids:

            i += 1
            sys.stdout.write('\rAsserting Avoid Clashes\t\t%i/%i' % (i, len(resources_ids)))
            resource = self.resources[resource_id]
            requiring_events = self.Xr[resource.num]

            for e1 in requiring_events:
                for e2 in requiring_events:
                    if e1 < e2:
                        self.optimizer.assert_expression(self.Xt[e1] & self.Xt[e2] == zero, weight)

        sys.stdout.write('\n')

    def preferTimesConstraint(self, cost, events_ids, times_ids, duration):

        sys.stdout.write('Asserting Prefer Times constraints')
        time_count = len(self.times)
        weight = cost/len(events_ids)

        # Create a list with all preferred time slots
        preferred = []
        for time_id in times_ids:
            preferred.append(self.times[time_id].num)

        mask = Const_BV(time_count, preferred)  # BitVector constant with all preferred bits set to '1'

        # for each event, forbid all forbidden times for the given duration (if any)
        for event_id in events_ids:
            event = self.events[event_id]
            if event.time is None:
                constraints = []
                e = event.num
                if duration is None:
                    for d in range(1, event.duration+1):
                        constraints.append(self.Xd[e][d] == self.Xd[e][d] & mask)
                    constraints.append(self.Xt[e] == self.Xt[e] & mask)            # Only neeeded constraint is this one (<--)
                    constraints.append(self.Xs[e] == self.Xs[e] & mask)            # All others are redundant
                elif duration in self.Xd[e]:
                    constraints.append(self.Xd[e][duration] == self.Xd[e][duration] & mask)
                self.optimizer.assert_expression(And(constraints), weight)

        sys.stdout.write('\t\t[DONE]\n')

    def splitEventsConstraint(self, cost, events_id, minimum, maximum, min_amount, max_amount):

        sys.stdout.write('Split Events constraints')

        Xs = self.Xs
        Xd = self.Xd
        time_count = len(self.times)
        weight = cost/len(events_id)

        zero = BitVecVal(0, time_count)
        for event_id in events_id:
            event = self.events[event_id]
            e = event.num
            constraints = []

            # nullify all durations below minimum
            for d in range(1, minimum):
                constraints.append(Xd[e][d] == zero)

            # nullify all durations above maximum
            for d in range(maximum+1, event.duration+1):
                constraints.append(Xd[e][d] == zero)

            if min_amount == max_amount:
                constraints.append(exactly_k_bits(Xs[e], min_amount))
            else:
                if min_amount > 1:
                    constraints.append(at_least_k_bits(Xs[e], min_amount))
                if max_amount < time_count:
                    constraints.append(at_most_k_bits(Xs[e], max_amount))
            self.optimizer.assert_expression(And(constraints), weight)

        sys.stdout.write('\t\t\t\t[DONE]\n')

    def spreadEventsConstraint(self, cost, event_groups, time_groups_ids):
        sys.stdout.write('Spread Events constraints')

        Xs = self.Xs
        Xt = self.Xt
        time_count = len(self.times)
        weight = cost/len(event_groups)

        for time_group_id, min_max in time_groups_ids.iteritems():
            time_group = self.time_groups[time_group_id]
            minimum = min_max[0]
            maximum = min_max[1]
            time_nums = []
            for time_id in time_group.elems:
                time_nums.append(self.times[time_id].num)
            time_nums.sort()
            # Generate a BitVector with the time_nums that do not have an immediate predecessor within the group
            heads_mask = Const_BV(time_count, [t for i, t in enumerate(time_nums) if t != time_nums[i-1]+1])
            # Generate a BitVector mask with all bits corresponding to times within time_group set
            mask = Const_BV(time_count, time_nums)

            for event_group_id in event_groups:
                event_group = self.event_groups[event_group_id]
                constraints = []

                # Z is the masked bitwise 'or' of Xs vectors of all group events
                Z = BitVec('Z_r_%i' % _get_unique_id(), time_count)
                z_args = []    # Xts will have a bit set iff one event of the group has its corresponding bit set in its Xt
                for event_id in event_group.elems:
                    event = self.events[event_id]
                    e = event.num
                    # All starting times of the time group are marked as such in Xs if they are set in Xt.
                    # that is, times from which its predecessor time is not within the time group,
                    constraints.append(Xs[e] == Xs[e] | (heads_mask & Xt[e]))
                    z_args.append(Xs[e])

                constraints.append(Z == (bv_or(z_args)) & mask)

                if minimum == maximum:
                    constraints.append(exactly_k_bits(Z, minimum))
                else:
                    # impose at_least minimum events from the group
                    if minimum > 0:
                        constraints.append(at_least_k_bits(Z, minimum))
                    # impose at_most maximum events from the group
                    if maximum < len(time_nums):
                        constraints.append(at_most_k_bits(Z, maximum))

                self.optimizer.assert_expression(And(constraints), weight)

        sys.stdout.write('\t\t\t\t[DONE]\n')

    def avoidUnavailableTimesConstraint(self, cost, resource_ids, time_ids):

        sys.stdout.write('\rAvoid Unavailable times constraints')

        time_count = len(self.times)
        weight = cost/len(resource_ids)

        # Create a mask with all unavailable times set to 0
        available_times = set(range(0, time_count))
        for time_id in time_ids:
            available_times.remove(self.times[time_id].num)
        mask = Const_BV(time_count, available_times)

        for resource_id in resource_ids:
            r = self.resources[resource_id].num

            # constraint the resource to attend events only at available times, that is,
            # constraint the events bound to this resource to occur only at available times
            for e in self.Xr[r]:
                constraints = [self.Xt[e] == self.Xt[e] & mask, self.Xs[e] == self.Xs[e] & mask]
                self.optimizer.assert_expression(And(constraints), weight)

        sys.stdout.write('\t\t[DONE]\n')

    def distributeSplitEventsConstraints(self, cost, events_ids, duration, minimum, maximum):

        sys.stdout.write('\rDistribute Split Events constraints')
        weight = cost/len(events_ids)

        for event_id in events_ids:
            event = self.events[event_id]

            if duration not in self.Xd[event.num]:
                raise Exception('invalid duration on distribute split events constraint for event: '
                                '%s (duration: %i)' % (event_id, duration))

            for i in range(1, minimum+1):
                self.optimizer.assert_expression(And(at_least_k_bits(self.Xd[event.num][duration], i)), weight)
            for i in range(maximum, event.duration):
                self.optimizer.assert_expression(And(at_most_k_bits(self.Xd[event.num][duration], i)), weight)

        sys.stdout.write('\t\t[DONE]\n')

    def limitIdleTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):

        sys.stdout.write('\rLimit Idle Times constraints')
        weight = cost/len(resource_ids)

        time_count = len(self.times)
        resource_nums = set()
        for resource_id in resource_ids:
            resource_nums.add(self.resources[resource_id].num)

        for time_group in time_groups:
            time_ids = self.time_groups[time_group].elems
            time_nums = set()
            for time_id in time_ids:
                time_nums.add(self.times[time_id].num)

            # Check compactness
            for t in range(min(time_nums)+1, max(time_nums)):
                if t not in time_nums:
                    raise Exception('Encountered a non compact TimeGroup in LimitIdleTimesConstraint')

            mask = Const_BV(time_count, time_nums)
            for r in resource_nums:
                constraints = []

                Rt = BitVec('Rt_%i__%i' % (r, _get_unique_id()), time_count)  # var with all busy times set
                Bf = BitVec('Bf_%i__%i' % (r, _get_unique_id()), time_count)  # mask for all times BEFORE first busy time
                Al = BitVec('Al_%i__%i' % (r, _get_unique_id()), time_count)  # mask for all times AFTER last busy time
                idle = BitVec('Idle_%i__%i' % (r, _get_unique_id()), time_count)  # var with all idle times set

                # set Rt
                rt_args = [self.Xt[e] for e in self.Xr[r]]
                bf_args = [(Rt << k) for k in range(1, len(time_nums))]
                al_args = [LShR(Rt, k) for k in range(1, len(time_nums))]

                # Fix Rt, Bf, Al and idle
                constraints.append(And(Rt == mask & (bv_or(rt_args)),
                                       Bf == mask & (bv_or(bf_args)),
                                       Al == mask & (bv_or(al_args)),
                                       idle == ~Rt & Bf & Al))

                self.optimizer.assert_expression(And(constraints))

                # Limit idle times
                # if minimum == maximum:
                #     constraints.append(exactly_k_bits(idle, minimum))
                # else:
                #     if minimum > 0:
                #         constraints.append(at_least_k_bits(idle, minimum))
                #     if maximum < len(time_nums)-2:
                #         constraints.append(at_most_k_bits(idle, maximum))

                # limit idle times
                for i in range(1, minimum+1):
                    self.optimizer.assert_expression(And(at_least_k_bits(idle, i)), weight)
                for i in range(maximum, len(time_groups)):
                    self.optimizer.assert_expression(And(at_most_k_bits(idle, i)), weight)

                # self.optimizer.assert_expression(And(constraints), weight)

        sys.stdout.write('\t\t\t[DONE]\n')

    def clusterBusyTimesConstraint(self, cost, resource_ids, time_groups, minimum, maximum):

        sys.stdout.write('\rCluster Busy Times constraint')
        weight = cost/len(resource_ids)

        time_count = len(self.times)
        resource_nums = set()
        for resource_id in resource_ids:
            resource_nums.add(self.resources[resource_id].num)

        for r in resource_nums:
            constraints = []

            Rt = BitVec('Rt_%i__%i' % (r, _get_unique_id()), time_count)  # var with all busy times set
            rt_args = [self.Xt[e] for e in self.Xr[r]]
            constraints.append(Rt == bv_or(rt_args))

            # var which will contain the first bit of each Timegroup set iff that TimeGroup is busy
            Btg = BitVec('Btg_%i__%i' % (r, _get_unique_id()), time_count)
            btg_args = []
            for time_group in time_groups:
                time_ids = self.time_groups[time_group].elems
                time_nums = set()
                for time_id in time_ids:
                    time_nums.add(self.times[time_id].num)

                first_bit_mask = Const_BV(time_count, [min(time_nums)])

                # var with the first bit of TimeGroup set iff the resource is busy at that TimeGroup
                Tg = BitVec('Tg_%i__%i' % (r, _get_unique_id()), time_count)
                tg_args = [(Rt << k) for k in range(0, max(time_nums)-min(time_nums)+1)]
                constraints.append(Tg == first_bit_mask & (bv_or(tg_args)))

                # Update Global busy time groups args
                btg_args.append(Tg)
            constraints.append(Btg == bv_or(btg_args))

            self.optimizer.assert_expression(And(constraints))

            # limit busy TimeGroups
            for i in range(1, minimum+1):
                self.optimizer.assert_expression(And(at_least_k_bits(Btg, i)), weight)
            for i in range(maximum, len(time_groups)):
                self.optimizer.assert_expression(And(at_most_k_bits(Btg, i)), weight)



        sys.stdout.write('\t\t\t[DONE]\n')

####################################################
#           BitVector encoding specifics           #
####################################################

def Const_BV(size, bits):
    num = 0
    for bit in bits:
        num += 2**(size-bit-1)
    return BitVecVal(num, size)

def weighted_cardinality(x, minimum, maximum, ub, weight, optimizer):
    zero = BitVecVal(0, x.size())
    op_id = _get_unique_id()
    # generate ub-1 auxiliary variables, each one will represent 'x' with 1 bit less
    bvs = [BitVec('Card_%i_%i' % (op_id, i), x.size()) for i in range(ub)]
    constraints = []
    act = x
    for i in range(ub):
        if i < minimum:
            optimizer.assert_expression(act != zero, weight)
        elif i >= maximum:
            optimizer.assert_expression(act == zero, weight)
        optimizer.assert_expression(bvs[i] == act & (act + BitVecVal(-1, x.size())))
        last = bvs[i]


def at_least_k_bits(x, k, last_var=False):
    """
    :param x: BitVector to constraint
    :param k: minimum number of bits that 'x' has to set
    :param last_var: indicates whether it should return a reference to the last auxiliary variable
    :return: z3 expression constraining 'x' to have at least 'k' bits set
    """
    zero = BitVecVal(0, x.size())
    if k > 1:
        y = BitVec('X_at_least__%i' % _get_unique_id(), x.size())
        constraint = And(y == x & x + BitVecVal(-1, x.size()), y != zero)  # y is x with one set bit "less"
        if last_var:
            expr, last = at_least_k_bits(y, k-1, last_var)
            if last is None:
                last = y
            return And(constraint, expr), last
        else:
            return And(constraint, at_least_k_bits(y, k-1))
    elif k == 1:
        if last_var:
            return x != zero, None
        else:
            return x != zero
    else:
        print 'warning: at_least_0??'

def at_most_k_bits(x, k):
    """
    :param x: BitVector to constraint
    :param k: maximum number of bits that 'x' has to set
    :return: z3 expression constraining 'x' to have at most 'k' bits set
    """
    zero = BitVecVal(0, x.size())
    if k > 0:
        y = BitVec('X_at_most__%i' % _get_unique_id(), x.size())
        constraint = y == x & x + BitVecVal(-1, x.size())  # y is x with one bit (set) "less"
        return And(constraint, Or(y == zero, at_most_k_bits(y, k-1)))
    elif k == 0:
        return x == zero
    else:
        print 'warning: at_most < 0??'

def exactly_k_bits(x, k):
    """
    :param x: BitVector to constraint
    :param k: number of bits that 'x' has to set
    :return: z3 expression constraining 'x' to have exactly 'k' bits set
    """
    zero = BitVecVal(0, x.size())
    if k > 1:
        # retrieve the last variable from the at_least_k, which represents x without k-1 set bits,
        # and impose that without another bit it equals to zero. (i.e. ensure that it only has 1 bit set)
        expr, y = at_least_k_bits(x, k, True)
        return And(expr, zero == y & y + BitVecVal(-1, x.size()))
    elif k == 1:
        return And(x != zero, zero == x & x + BitVecVal(-1, x.size()))
    elif k == 0:
        return x == zero
    else:
        print 'warning: exactly < 0??'

def bv_and(ls):
    if len(ls) == 1:
        return ls[0]
    elif len(ls) == 2:
        return ls[0] & ls[1]
    elif len(ls) > 2:
        return ls[0] & bv_and(ls[1:])
    else:
        raise Exception('Invalid number of arguments')

def bv_or(ls):
    if len(ls) == 1:
        return ls[0]
    elif len(ls) == 2:
        return ls[0] | ls[1]
    elif len(ls) > 2:
        return ls[0] | bv_or(ls[1:])
    else:
        raise Exception('Invalid number of arguments')

def _get_unique_id():
    global _unique_id
    _unique_id += 1
    return _unique_id-1
































