"""
 Copyright 2015 Cristofor Nogueira Gascons
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import time

__author__ = 'cris Nogueira'

from z3 import *
import random

class Optimizer:
    """
    This class will hold a reference to a z3 solver (either provided or a default general purpose one)
    And a list holding z3 ast expressions and labeled with a natural integer representing its priority
    being infinity top priority (a.k.a hard-constraint) and 0 the least of the priorities
    """

    __NUMBER = 1

    def __init__(self, solver=Solver()):

        self.solver = solver
        self.soft_constraints = []  # list of tuples (Protection variable, weight)
        self.op = Operation()

    def __del__(self):
        self.solver = None
        self.soft_constraints = None

    def assert_expression(self, expression, weight=float('inf')):
        if weight == float('inf'):
            self.solver.assert_exprs(expression)
        else:
            p = Bool('p_%i' % self.__NUMBER)
            self.__NUMBER += 1
            self.solver.assert_exprs(Implies(p, expression))
            self.soft_constraints.append((p, weight))

    def binary_optimization(self, timeout=300000):
        
        timeout_param = 'timeout'
        try:
            self.solver.set(timeout_param, timeout)
        except:
            timeout_param = 'soft-timeout'
            self.solver.set(timeout_param, timeout)
        # first check if hard-constraints are satisfiable
        res = self.solver.check([Not(p) for p, w in self.soft_constraints])
        if res == sat:
            start = time.time()
            model = self.solver.model()
            lower_bound = 0
            upper_bound = sum([w for p, w in self.soft_constraints])
            while lower_bound < upper_bound:
                k = (lower_bound + upper_bound) / 2
                # check if the soft soft constraints can have a cost lower or equal than k
                self.solver.push()
                self.solver.assert_exprs(Sum([If(p, 0, w) for p, w in self.soft_constraints]) <= k)
                remaining_timeout = int(timeout-1000*(time.time()-start))
                if remaining_timeout <= 0:
                    break
                self.solver.set(timeout_param, remaining_timeout)
                res = self.solver.check()
                if res == sat:
                    upper_bound = k
                    model = self.solver.model()
                    print 'new optimum found\t\tk = %i    \t\tLB = %i\tUB = %i\ttime mark: %f seconds' % \
                          (k, lower_bound, upper_bound, time.time()-start)
                elif res == unsat:
                    lower_bound = k+1
                    print '\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tLB = %i\tUB = %i\ttime mark: %f seconds' % \
                          (lower_bound, upper_bound, time.time()-start)
                    self.solver.pop()
                else:
                    # timed out
                    print 'timed out'
                    break
            return model
        elif res == unknown:
            reason = self.solver.reason_unknown()
            print reason
        else:
            self.op.on_unsat()
            return None

class Operation:

    def __init__(self):
        pass

    def on_sat(self, solver, cost):
        print '\tsat ->', cost

    def on_unsat(self):
        print 'unsat'

    def on_unknown(self, solver, cost):
        print '\tunknown ->', cost

