"""
 Copyright 2015 Cristofor Nogueira Gascons
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = 'cris nogueira'

import xml.etree.ElementTree

class MalformedXHSTT(Exception):

    def __init__(self, message='Error: given XML does not follow XHSTT convention. Aborting.'):
        self.message = message

    def __str__(self):
        return self.message

class Parser:

    def __init__(self, model, xml_source):
        self.model = model
        self.xml_source = None
        self.xml_root = None
        self.set_source_xml(xml_source)
        self.assign_time = set()
        self.assign_res = dict()

    def __str__(self):
        if not self.xml_source:
            return 'undefined xml_source'
        else:
            return str(self.xml_source)

    def set_source_xml(self, source_xml):
        self.xml_source = source_xml
        self.xml_root = xml.etree.ElementTree.parse(source_xml).getroot()

    def parse_model(self, source_xml=None):
        """
        Parses all needed elems to create a model without restrictions (just assign time & resources taken into account)
        """
        if source_xml is None:
            source_xml = self.xml_source

        self.set_source_xml(source_xml)

        # Ignore all child tags but <Instances> and just parse the first instance below that
        xml_instance = self.xml_root.find('Instances').find('Instance')
        self.__parse_instance(xml_instance)

        # Now parse assign times & resources constraints (if any)
        # Variables assignment
        xml_node = xml_instance.find('Constraints')

    def __parse_instance(self, xml_instance):

        if xml_instance is None:
            raise MalformedXHSTT()

        # Parse Times
        self.__parse_times(xml_instance.find('Times'))

        # Parse Resources
        self.__parse_resources(xml_instance.find('Resources'))

        # Parse Events
        self.__parse_events(xml_instance.find('Events'))

        # Parse Constraints
        self.__parse_constraints()

    def __parse_assign_time_constraint(self, xml_node):
        events = set()
        if xml_node is not None and self.__constraint_required(xml_node):
            xml_apply = xml_node.find('AppliesTo')

            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))

            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

        # return all involved events
        return events

    def __parse_assign_resource_constraint(self, xml_node):

        if xml_node is not None and self.__constraint_required(xml_node):
            xml_apply = xml_node.find('AppliesTo')

            events = set()
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))

            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

            role = xml_node.find('Role').text

            # Once retrieved all involved events and role, annotate it on a private variable
            for event in events:
                if event not in self.assign_res:
                    self.assign_res[event] = set()
                self.assign_res[event].add(role)

    def __parse_time_groups(self, xml_groups):

        # xml_times IS optional
        if xml_groups is not None:

            for xml_group in xml_groups.findall('Week'):
                self.model.declare_time_group(xml_group.get('Id'), xml_group.find('Name').text, 'Week')

            for xml_group in xml_groups.findall('Day'):
                self.model.declare_time_group(xml_group.get('Id'), xml_group.find('Name').text, 'Day')

            for xml_group in xml_groups.findall('TimeGroup'):
                self.model.declare_time_group(xml_group.get('Id'), xml_group.find('Name').text)

    def __parse_times(self, xml_times):

        if xml_times is None:
            raise MalformedXHSTT

        # Parse time groups
        self.__parse_time_groups(xml_times.find('TimeGroups'))

        # Iterate over all <Time> tags, updating each one into our given model
        for xml_elem in xml_times.findall('Time'):

            # Get element id
            identifier = xml_elem.get('Id')

            # Register the time element in the model
            self.model.register_time(identifier, xml_elem.find('Name').text)

            xml_week = xml_elem.find('Week')
            if xml_week is not None:
                self.model.time_to_group(identifier, xml_week.get('Reference'))

            xml_day = xml_elem.find('Day')
            if xml_day is not None:
                self.model.time_to_group(identifier, xml_day.get('Reference'))

            xml_groups = xml_elem.find('TimeGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('TimeGroup'):
                    self.model.time_to_group(identifier, xml_group.get('Reference'))


    def __parse_resource_types(self, xml_resources):
        if xml_resources is None:
            print 'warning, given XHSTT has no resource types in it'
        else:
            for xml_rtype in xml_resources.findall('ResourceType'):
                self.model.register_resource_type(xml_rtype.get('Id'), xml_rtype.find('Name').text)

    def __parse_resource_groups(self, xml_rgroups):
        if xml_rgroups is not None:
            for xml_rgroup in xml_rgroups:
                identifier = xml_rgroup.get('Id')
                name = xml_rgroup.find('Name').text
                rtype = xml_rgroup.find('ResourceType').get('Reference')
                self.model.declare_resource_group(identifier, name, rtype)

    def __parse_resources(self, xml_resources):

        if xml_resources is None:
            raise MalformedXHSTT

        # Self explanatory.
        self.__parse_resource_types(xml_resources.find('ResourceTypes'))

        # Parse resource groups
        self.__parse_resource_groups(xml_resources.find('ResourceGroups'))

        # Iterate over all <Resource> tags, updating each one into our given model
        for xml_elem in xml_resources.findall('Resource'):

            # Get element id
            identifier = xml_elem.get('Id')
            name = xml_elem.find('Name').text
            rtype = xml_elem.find('ResourceType').get('Reference')

            # Register the Resource element in the model
            self.model.register_resource(identifier, name, rtype)

            # parse the groups this element belongs to
            xml_groups = xml_elem.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    self.model.resource_to_group(identifier, xml_group.get('Reference'))

    def __parse_optimize_events(self):

        return

        xml_node = self.xml_root.find('SolutionGroups')
        xml_node = xml_node[-1].find('Solution').find('Events')

        n = int(len(xml_node)*0.15)
        print n
        for i in range(n):
            xml_act = xml_node[(i*n) % len(xml_node)]

            event_num = self.model.get_event_by_ref(xml_act.get('Reference')).num
            duration = int(xml_act.find('Duration').text)
            start_t = self.model.get_time_by_ref(xml_act.find('Time').get('Reference')).num

            self.model.set_time_for_event(event_num, duration, start_t)

    def __parse_event_groups(self, xml_egroups):
        if xml_egroups is not None:
            # parse all courses
            for xml_course in xml_egroups.findall('Course'):
                self.model.declare_event_group(xml_course.get('Id'), xml_course.find('Name').text, 'Course')
            # parse rest of event groups
            for xml_egroup in xml_egroups.findall('EventGroup'):
                self.model.declare_event_group(xml_egroup.get('Id'), xml_egroup.find('Name').text)

    def __parse_events(self, xml_events):

        if xml_events is None:
            raise MalformedXHSTT

        # Parse Event Groups (if any)
        self.__parse_event_groups(xml_events.find('EventGroups'))

        # Pre-parse all events (just to fill event groups)
        for xml_elem in xml_events.findall('Event'):

            # Get element id
            identifier = xml_elem.get('Id')
            color = xml_elem.get('Color')
            name = xml_elem.find('Name').text
            duration = int(xml_elem.find('Duration').text)

            # Register the Event element in the model
            self.model.register_event(identifier, name, duration, color)

            # Parse Event groups attached to the event
            xml_course = xml_elem.find('Course')
            if xml_course is not None:
                self.model.event_to_group(identifier, xml_course.get('Reference'))
            xml_groups = xml_elem.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    self.model.event_to_group(identifier, xml_group.get('Reference'))

        # parse all assign-like constraints in order to be able to filter incoming events
        xml_constraints = self.xml_root.find('Instances').find('Instance').find('Constraints')
        events = set()
        for xml_constraint in xml_constraints.findall('AssignTimeConstraint'):
            events.update(self.__parse_assign_time_constraint(xml_constraint))
        # open assign time set
        self.assign_time = events
        for xml_constraint in xml_constraints.findall('AssignResourceConstraint'):
            self.__parse_assign_resource_constraint(xml_constraint)

        # Iterate over all <Event> tags, updating each one into our given model
        for xml_elem in xml_events.findall('Event'):

            # Get element id
            identifier = xml_elem.get('Id')

            # update all its resource requirements
            event = self.model.get_event_by_ref(identifier)
            xml_resources = xml_elem.find('Resources')
            if xml_resources is None or xml_resources.find('Resource') is None:
                raise Exception('Unsupported feature: Encountered an Event without any resource requirements')
            else:
                for xml_resource in xml_resources.findall('Resource'):

                    reference = xml_resource.get('Reference')

                    resource = self.model.get_resource_by_ref(reference)
                    role = None
                    rtype = None

                    if xml_resource.find('Role') is not None:
                        role = xml_resource.find('Role').text

                    if xml_resource.find('ResourceType') is not None:
                        rtype = xml_resource.find('ResourceType').get('Reference')

                    # If event has no preassigned resource for a given role, check that it is intended to be filled
                    if reference is None and (identifier not in self.assign_res or role not in self.assign_res[identifier]):
                        print 'Warning: role \'%s\' for event \'%s\' is not intended to be filled by the model' % (role, identifier)
                    else:
                        # update the event with the parsed requirement
                        event.attach_resource(resource, role, rtype, self.model)

            # Check if the event is expected to attend a preassigned time
            xml_time = xml_elem.find('Time')
            if xml_time is not None:
                ref = xml_time.get('Reference')
                if self.model.get_time_by_ref(ref) is not None:
                    event.set_time(ref)
                else:
                    raise Exception('Event with an undeclared Time reference')
            # check if the model specifies that a a time has to be assigned to the event
            elif identifier not in self.assign_time:
                raise Exception('Event without Time nor assign time constraint')

        # notify the model that all events have been fully parsed (before any constraints ofc)
        self.model.onParsedEvents(xml_constraints.find('SpreadEventsConstraint') is not None)

        self.__parse_optimize_events()
        ####################################
        #            CONSTRAINTS           #
        ####################################


    def __parse_prefer_resources_constraint(self, xml_node):

        if xml_node is not None and self.__constraint_required(xml_node):
            xml_apply = xml_node.find('AppliesTo')

            # Gather all events to which this constraint applies
            events = set()
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))
            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

            # Gather all resources to which this constraint refers
            resources = set()
            xml_groups = xml_node.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    resources.update(self.model.get_resources_from_group(xml_group.get('Reference')))
            xml_resources = xml_node.find('Resources')
            if xml_resources is not None:
                for xml_resource in xml_resources.findall('Resource'):
                    resources.add(xml_resource.get('Reference'))

            role = xml_node.find('Role').text

            # propagate the parsed info to the model
            self.model.preferResourcesConstraint(events, resources, role)

    def __parse_prefer_times_constraint(self, xml_node):

        if xml_node is not None:
            xml_apply = xml_node.find('AppliesTo')

            # Gather all events to which this constraint applies
            events = set()
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))
            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

            # Gather all times to which this constraint refers
            times = set()
            xml_groups = xml_node.find('TimeGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('TimeGroup'):
                    times.update(self.model.get_times_from_group(xml_group.get('Reference')))
            xml_times = xml_node.find('Times')
            if xml_times is not None:
                for xml_time in xml_times.findall('Time'):
                    times.add(xml_time.get('Reference'))

            # Gather duration of the events to which this constraint applies
            xml_duration = xml_node.find('Duration')
            duration = None
            if xml_duration is not None:
                duration = int(xml_duration.text)

            ls = [e for e in events if not self.model.get_event_by_ref(e).hasPreassignedTime()]
            pointsOfApp = len(ls)
            cost = self.__constraint_cost(xml_node, pointsOfApp)

            # propagate the parsed info to the model
            self.model.preferTimesConstraint(cost, events, times, duration)

    def __constraint_cost(self, xml_node, nPointsOfApplication):
        required = 'true' == xml_node.find('Required').text.lower()
        if required:
            return float('inf')

        weight = int(xml_node.find('Weight').text)
        costFunc = xml_node.find('CostFunction').text.lower()
        if costFunc in ['sum', 'linear']:
            return weight*nPointsOfApplication
        elif costFunc in ['quadratic']:
            return weight*(nPointsOfApplication**2)
        elif costFunc in ['step']:
            return weight*int(nPointsOfApplication > 0)
        else:
            print 'unrecognized cost function: ', xml_node.find('CostFunction').text


    def __parse_avoid_clashes_constraint(self, xml_node):
        if xml_node is not None:
            xml_apply = xml_node.find('AppliesTo')

            # Gather all resources to which this constraint refers
            resources = set()
            xml_groups = xml_apply.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    resources.update(self.model.get_resources_from_group(xml_group.get('Reference')))
            xml_resources = xml_apply.find('Resources')
            if xml_resources is not None:
                for xml_resource in xml_resources.findall('Resource'):
                    resources.add(xml_resource.get('Reference'))

            pointsOfApp = len(resources)
            cost = self.__constraint_cost(xml_node, pointsOfApp)

            # propagate the parsed info to the model
            self.model.avoidClashesConstraint(cost, resources)

    def __parse_split_events_constraint(self, xml_node):
        if xml_node is not None:

            # Gather all events to which this constraint applies
            events = set()
            xml_apply = xml_node.find('AppliesTo')
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))
            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

            minimum = int(xml_node.find('MinimumDuration').text)
            maximum = int(xml_node.find('MaximumDuration').text)
            min_amount = int(xml_node.find('MinimumAmount').text)
            max_amount = int(xml_node.find('MaximumAmount').text)

            # Check integrity
            if  maximum < minimum or max_amount < min_amount or not events:
                raise Exception('Failed integrity check of split events constraint')

            pointsOfApp = len(events)
            cost = self.__constraint_cost(xml_node, pointsOfApp)

            self.model.splitEventsConstraint(cost, events, minimum, maximum, min_amount, max_amount)

    def __parse_spread_events_constraint(self, xml_node):
        if xml_node is not None:

            # Gather all event_groups to which this constraint applies
            events_groups = set()
            xml_apply = xml_node.find('AppliesTo')
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events_groups.add(xml_group.get('Reference'))

            # Gather all time groups to which this constraints applies
            time_groups = dict()
            xml_time_groups = xml_node.find('TimeGroups')
            for xml_time_group in xml_time_groups.findall('TimeGroup'):

                reference = xml_time_group.get('Reference')
                minimum = int(xml_time_group.find('Minimum').text)
                maximum = int(xml_time_group.find('Maximum').text)

                time_groups[reference] = (minimum, maximum)

            pointsOfApp = len(events_groups)
            cost = self.__constraint_cost(xml_node, pointsOfApp)

            self.model.spreadEventsConstraint(cost, events_groups, time_groups)

    def __parse_avoid_unavailable_times_constraint(self, xml_node):
        if xml_node is not None:

            # Gather all resources to which this constraint refers
            resources = set()
            xml_apply = xml_node.find('AppliesTo')
            xml_groups = xml_apply.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    resources.update(self.model.get_resources_from_group(xml_group.get('Reference')))
            xml_resources = xml_apply.find('Resources')
            if xml_resources is not None:
                for xml_resource in xml_resources.findall('Resource'):
                    resources.add(xml_resource.get('Reference'))

            times = set()
            xml_groups = xml_node.find('TimeGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('TimeGroup'):
                    times.update(self.model.get_times_from_group(xml_group))
            xml_times = xml_node.find('Times')
            if xml_times is not None:
                for xml_time in xml_times.findall('Time'):
                    times.add(xml_time.get('Reference'))

            pointsOfApp = len(resources)
            cost = self.__constraint_cost(xml_node, pointsOfApp)

            self.model.avoidUnavailableTimesConstraint(cost, resources, times)

    def __parse_distribute_split_events_constraint(self, xml_constraint):
        if xml_constraint is not None:

            events = set()
            xml_apply = xml_constraint.find('AppliesTo')
            xml_groups = xml_apply.find('EventGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('EventGroup'):
                    events.update(self.model.get_events_from_group(xml_group.get('Reference')))

            xml_events = xml_apply.find('Events')
            if xml_events is not None:
                for xml_event in xml_events.findall('Event'):
                    events.add(xml_event.get('Reference'))

            duration = int(xml_constraint.find('Duration').text)
            minimum = int(xml_constraint.find('Minimum').text)
            maximum = int(xml_constraint.find('Maximum').text)

             # Check integrity
            if maximum < minimum or not events:
                raise Exception('Failed integrity check of distribute split events constraint')

            pointsOfApp = len(events)
            cost = self.__constraint_cost(xml_constraint, pointsOfApp)

            self.model.distributeSplitEventsConstraints(cost, events, duration, minimum, maximum)

    def __parse_limit_idle_times_constraint(self, xml_constraint):
        if xml_constraint is not None:

            # Gather all resources to which this constraint refers
            xml_apply = xml_constraint.find('AppliesTo')
            resources = set()
            xml_groups = xml_apply.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    resources.update(self.model.get_resources_from_group(xml_group.get('Reference')))
            xml_resources = xml_apply.find('Resources')
            if xml_resources is not None:
                for xml_resource in xml_resources.findall('Resource'):
                    resources.add(xml_resource.get('Reference'))

            time_groups = set()
            for xml_time_group in xml_constraint.find('TimeGroups').findall('TimeGroup'):
                time_groups.add(xml_time_group.get('Reference'))

            minimum = int(xml_constraint.find('Minimum').text)
            maximum = int(xml_constraint.find('Maximum').text)

            pointsOfApp = len(resources)
            cost = self.__constraint_cost(xml_constraint, pointsOfApp)

            self.model.limitIdleTimesConstraint(cost, resources, time_groups, minimum, maximum)

    def __parse_cluster_busy_times_constraint(self, xml_constraint):
        if xml_constraint is not None:

            # Gather all resources to which this constraint refers
            xml_apply = xml_constraint.find('AppliesTo')
            resources = set()
            xml_groups = xml_apply.find('ResourceGroups')
            if xml_groups is not None:
                for xml_group in xml_groups.findall('ResourceGroup'):
                    resources.update(self.model.get_resources_from_group(xml_group.get('Reference')))
            xml_resources = xml_apply.find('Resources')
            if xml_resources is not None:
                for xml_resource in xml_resources.findall('Resource'):
                    resources.add(xml_resource.get('Reference'))

            time_groups = set()
            for xml_time_group in xml_constraint.find('TimeGroups').findall('TimeGroup'):
                time_groups.add(xml_time_group.get('Reference'))

            minimum = int(xml_constraint.find('Minimum').text)
            maximum = int(xml_constraint.find('Maximum').text)

            pointsOfApp = len(resources)
            cost = self.__constraint_cost(xml_constraint, pointsOfApp)

            self.model.clusterBusyTimesConstraint(cost, resources, time_groups, minimum, maximum)

    def __parse_constraints(self):

        # Ignore all child tags but <Instances> and just parse the first instance below that
        xml_constraints = self.xml_root.find('Instances').find('Instance').find('Constraints')

        # Prefer-like constraints
        for xml_constraint in xml_constraints.findall('PreferResourcesConstraint'):
            self.__parse_prefer_resources_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('AvoidClashesConstraint'):
            self.__parse_avoid_clashes_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('SplitEventsConstraint'):
            self.__parse_split_events_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('SpreadEventsConstraint'):
            self.__parse_spread_events_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('PreferTimesConstraint'):
            self.__parse_prefer_times_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('AvoidUnavailableTimesConstraint'):
            self.__parse_avoid_unavailable_times_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('DistributeSplitEventsConstraint'):
            self.__parse_distribute_split_events_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('LimitIdleTimesConstraint'):
            self.__parse_limit_idle_times_constraint(xml_constraint)

        for xml_constraint in xml_constraints.findall('ClusterBusyTimesConstraint'):
            self.__parse_cluster_busy_times_constraint(xml_constraint)

        #    TODO


    def __constraint_required(self, xml_constraint):
        return xml_constraint.find('Required').text == 'true'











